# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import re


PUBLIC_PREFIX = "public."
GLYPH_ORDER_KEY = PUBLIC_PREFIX + "glyphOrder"
OBJECT_LIBS_KEY = PUBLIC_PREFIX + "objectLibs"

GLYPHS_PREFIX = "com.schriftgestaltung."
GLYPHLIB_PREFIX = GLYPHS_PREFIX + "Glyphs."
ROBOFONT_PREFIX = "com.typemytype.robofont."
UFO2FT_FILTERS_KEY = "com.github.googlei18n.ufo2ft.filters"
UFO2FT_USE_PROD_NAMES_KEY = "com.github.googlei18n.ufo2ft.useProductionNames"
ANNOTATIONS_LIB_KEY = GLYPHS_PREFIX + "annotations"
COMPONENT_INFO_KEY = GLYPHLIB_PREFIX + "ComponentInfo"
UFO_FILENAME_CUSTOM_PARAM = "UFO Filename"

BACKGROUND_IMAGE_PREFIX = GLYPHS_PREFIX + "backgroundImage."
CROP_KEY = BACKGROUND_IMAGE_PREFIX + "crop"
LOCKED_KEY = BACKGROUND_IMAGE_PREFIX + "locked"
ALPHA_KEY = BACKGROUND_IMAGE_PREFIX + "alpha"

BRACKET_GLYPH_TEMPLATE = "{glyph_name}.BRACKET.{description}"
BRACKET_GLYPH_RE = re.compile(r"(?P<glyph_name>.+)\.BRACKET.(?P<box>.*)$")
BRACKET_GLYPH_SUFFIX_RE = re.compile(r".*(\.BRACKET\..*)$")

MASTER_CUSTOM_PARAM_PREFIX = GLYPHS_PREFIX + "customParameter.GSFontMaster."
FONT_CUSTOM_PARAM_PREFIX = GLYPHS_PREFIX + "customParameter.GSFont."

ANONYMOUS_FEATURE_PREFIX_NAME = "<anonymous>"
ORIGINAL_FEATURE_CODE_KEY = GLYPHLIB_PREFIX + "originalFeatureCode"
ORIGINAL_CATEGORY_KEY = GLYPHLIB_PREFIX + "originalOpenTypeCategory"

INSERT_FEATURE_MARKER_RE = r"\s*# Automatic Code.*"
INSERT_FEATURE_MARKER_COMMENT = "# Automatic Code\n"

APP_VERSION_LIB_KEY = GLYPHS_PREFIX + "appVersion"
KEYBOARD_INCREMENT_KEY = GLYPHS_PREFIX + "keyboardIncrement"
MASTER_ORDER_LIB_KEY = GLYPHS_PREFIX + "fontMasterOrder"

SCRIPT_LIB_KEY = GLYPHLIB_PREFIX + "script"
ORIGINAL_WIDTH_KEY = GLYPHLIB_PREFIX + "originalWidth"
BACKGROUND_WIDTH_KEY = GLYPHLIB_PREFIX + "backgroundWidth"

UFO_ORIGINAL_KERNING_GROUPS_KEY = GLYPHLIB_PREFIX + "originalKerningGroups"
UFO_GROUPS_NOT_IN_FEATURE_KEY = GLYPHLIB_PREFIX + "groupsNotInFeature"
UFO_KERN_GROUP_PATTERN = re.compile("^public\\.kern([12])\\.(.*)$")

LOCKED_GUIDE_NAME_SUFFIX = " [locked]"

HINTS_LIB_KEY = GLYPHS_PREFIX + "hints"
SHAPE_ORDER_LIB_KEY = GLYPHLIB_PREFIX + "shapeOrder"

SMART_COMPONENT_AXES_LIB_KEY = GLYPHS_PREFIX + "smartComponentAxes"

EXPORT_KEY = GLYPHS_PREFIX + "export"
WIDTH_KEY = GLYPHS_PREFIX + "width"
WEIGHT_KEY = GLYPHS_PREFIX + "weight"
FULL_FILENAME_KEY = GLYPHLIB_PREFIX + "fullFilename"
MANUAL_INTERPOLATION_KEY = GLYPHS_PREFIX + "manualInterpolation"
# Following typo kept for backwards compatibility
INSTANCE_INTERPOLATIONS_KEY = GLYPHS_PREFIX + "intanceInterpolations"

CUSTOM_PARAMETERS_KEY = GLYPHS_PREFIX + "customParameters"
CUSTOM_PARAMETERS_BLACKLIST = [
    # These are stored in the official descriptor attributes.
    "familyName",
    "postscriptFontName",
    "fileName",
    # These can be recovered by reading the mapping backward.
    "weightClass",
    "widthClass",
    # These are artificial.
    FULL_FILENAME_KEY,
    UFO_FILENAME_CUSTOM_PARAM,
]

# Reference:
# https://github.com/googlefonts/glyphsLib/pull/881#issuecomment-1474226616
PROPERTIES_KEY = GLYPHS_PREFIX + "properties"
PROPERTIES_WHITELIST = [
    # This is stored in the official descriptor attributes.
    # "familyNames",
    "designers",
    "designerURL",
    "manufacturers",
    "manufacturerURL",
    "copyrights",
    "versionString",
    "vendorID",
    "uniqueID",
    "licenses",
    "licenseURL",
    "trademarks",
    "descriptions",
    "sampleTexts",
    "postscriptFullNames",
    "postscriptFullName",
    # This is stored in the official descriptor attributes.
    # "postscriptFontName",
    "compatibleFullNames",
    "styleNames",
    "styleMapFamilyNames",
    "styleMapStyleNames",
    "preferredFamilyNames",
    "preferredSubfamilyNames",
    "variableStyleNames",
    "WWSFamilyName",
    "WWSSubfamilyName",
    "variationsPostScriptNamePrefix",
]

LAYER_ID_KEY = GLYPHS_PREFIX + "layerId"
LAYER_ORDER_PREFIX = GLYPHS_PREFIX + "layerOrderInGlyph."
LAYER_ORDER_TEMP_USER_DATA_KEY = "__layerOrder"

MASTER_ID_LIB_KEY = GLYPHS_PREFIX + "fontMasterID"
UFO_FILENAME_KEY = GLYPHLIB_PREFIX + "ufoFilename"
UFO_YEAR_KEY = GLYPHLIB_PREFIX + "ufoYear"
UFO_NOTE_KEY = GLYPHLIB_PREFIX + "ufoNote"

UFO_DATA_KEY = GLYPHLIB_PREFIX + "ufoData"
FONT_USER_DATA_KEY = GLYPHLIB_PREFIX + "fontUserData"
LAYER_LIB_KEY = GLYPHLIB_PREFIX + "layerLib"
LAYER_NAME_KEY = GLYPHLIB_PREFIX + "layerName"
GLYPH_USER_DATA_KEY = GLYPHLIB_PREFIX + "glyphUserData"
NODE_USER_DATA_KEY = GLYPHLIB_PREFIX + "nodeUserData"


GLYPHS_COLORS = (
    "0.85,0.26,0.06,1",
    "0.99,0.62,0.11,1",
    "0.65,0.48,0.2,1",
    "0.97,1,0,1",
    "0.67,0.95,0.38,1",
    "0.04,0.57,0.04,1",
    "0,0.67,0.91,1",
    "0.18,0.16,0.78,1",
    "0.5,0.09,0.79,1",
    "0.98,0.36,0.67,1",
    "0.75,0.75,0.75,1",
    "0.25,0.25,0.25,1",
)

# https://www.microsoft.com/typography/otspec/os2.htm#cpr
CODEPAGE_RANGES = {
    1252: 0,
    1250: 1,
    1251: 2,
    1253: 3,
    1254: 4,
    1255: 5,
    1256: 6,
    1257: 7,
    1258: 8,
    # 9-15: Reserved for Alternate ANSI
    874: 16,
    932: 17,
    936: 18,
    949: 19,
    950: 20,
    1361: 21,
    # 22-28: Reserved for Alternate ANSI and OEM
    # 29: Macintosh Character Set (US Roman)
    # 30: OEM Character Set
    # 31: Symbol Character Set
    # 32-47: Reserved for OEM
    869: 48,
    866: 49,
    865: 50,
    864: 51,
    863: 52,
    862: 53,
    861: 54,
    860: 55,
    857: 56,
    855: 57,
    852: 58,
    775: 59,
    737: 60,
    708: 61,
    850: 62,
    437: 63,
}

REVERSE_CODEPAGE_RANGES = {value: key for key, value in CODEPAGE_RANGES.items()}

UFO2FT_FEATURE_WRITERS_KEY = "com.github.googlei18n.ufo2ft.featureWriters"

UFO2FT_COLOR_PALETTES_KEY = "com.github.googlei18n.ufo2ft.colorPalettes"
UFO2FT_COLOR_LAYER_MAPPING_KEY = "com.github.googlei18n.ufo2ft.colorLayerMapping"
UFO2FT_COLOR_LAYERS_KEY = "com.github.googlei18n.ufo2ft.colorLayers"

UFO2FT_META_TABLE_KEY = PUBLIC_PREFIX + "openTypeMeta"

DEFAULT_FEATURE_WRITERS = [
    {"class": "CursFeatureWriter"},
    {"class": "KernFeatureWriter"},
    {"class": "MarkFeatureWriter"},
    {"class": "GdefFeatureWriter"},
]

DEFAULT_LAYER_NAME = PUBLIC_PREFIX + "default"

# From the spec:
# https://docs.microsoft.com/en-gb/typography/opentype/spec/os2#uswidthclass
WIDTH_CLASS_TO_VALUE = {
    1: 50,  # Ultra-condensed
    2: 62.5,  # Extra-condensed
    3: 75,  # Condensed
    4: 87.5,  # Semi-condensed
    5: 100,  # Medium
    6: 112.5,  # Semi-expanded
    7: 125,  # Expanded
    8: 150,  # Extra-expanded
    9: 200,  # Ultra-expanded
}

LANGUAGE_MAPPING = {
    "dflt": None,
    "AFK": 0x0436,
    "ARA": 0x0C01,
    "ASM": 0x044D,
    "AZE": 0x042C,
    "BEL": 0x0423,
    "BEN": 0x0845,
    "BGR": 0x0402,
    "BRE": 0x047E,
    "CAT": 0x0403,
    "CSY": 0x0405,
    "DAN": 0x0406,
    "DEU": 0x0407,
    "ELL": 0x0408,
    "ENG": 0x0409,
    "ESP": 0x0C0A,
    "ETI": 0x0425,
    "EUQ": 0x042D,
    "FIN": 0x040B,
    "FLE": 0x0813,
    "FOS": 0x0438,
    "FRA": 0x040C,
    "FRI": 0x0462,
    "GRN": 0x046F,
    "GUJ": 0x0447,
    "HAU": 0x0468,
    "HIN": 0x0439,
    "HRV": 0x041A,
    "HUN": 0x040E,
    "HVE": 0x042B,
    "IRI": 0x083C,
    "ISL": 0x040F,
    "ITA": 0x0410,
    "IWR": 0x040D,
    "JPN": 0x0411,
    "KAN": 0x044B,
    "KAT": 0x0437,
    "KAZ": 0x043F,
    "KHM": 0x0453,
    "KOK": 0x0457,
    "KOR": 0x0412,
    "LAO": 0x0454,
    "LSB": 0x082E,
    "LTH": 0x0427,
    "LVI": 0x0426,
    "MAR": 0x044E,
    "MKD": 0x042F,
    "MLR": 0x044C,
    "MLY": 0x043E,
    "MOL": 0x0818,
    "MNG": 0x0352,
    "MTS": 0x043A,
    "NEP": 0x0461,
    "NLD": 0x0413,
    "NOB": 0x0414,
    "ORI": 0x0448,
    "PAN": 0x0446,
    "PAS": 0x0463,
    "PLK": 0x0415,
    "PTG": 0x0816,
    "PTG-BR": 0x0416,
    "RMS": 0x0417,
    "ROM": 0x0418,
    "RUS": 0x0419,
    "SAN": 0x044F,
    "SKY": 0x041B,
    "SLV": 0x0424,
    "SQI": 0x041C,
    "SRB": 0x081A,
    "SVE": 0x041D,
    "TAM": 0x0449,
    "TAT": 0x0444,
    "TEL": 0x044A,
    "THA": 0x041E,
    "TIB": 0x0451,
    "TRK": 0x041F,
    "UKR": 0x0422,
    "URD": 0x0420,
    "USB": 0x042E,
    "UYG": 0x0480,
    "UZB": 0x0443,
    "VIT": 0x042A,
    "WEL": 0x0452,
    "ZHH": 0x0C04,
    "ZHS": 0x0804,
    "ZHT": 0x0404,
}

REVERSE_LANGUAGE_MAPPING = {v: k for k, v in LANGUAGE_MAPPING.items()}

GLYPHS_MATH_PREFIX = "com.nagwa.MATHPlugin."
GLYPHS_MATH_CONSTANTS_KEY = GLYPHS_MATH_PREFIX + "constants"
GLYPHS_MATH_VARIANTS_KEY = GLYPHS_MATH_PREFIX + "variants"
GLYPHS_MATH_EXTENDED_SHAPE_KEY = GLYPHS_MATH_PREFIX + "extendedShape"



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import argparse
import os
import sys

import glyphsLib


def main(args=None):
    if not args:
        args = sys.argv[1:]

    python_executable = os.path.basename(sys.executable)
    parser = argparse.ArgumentParser(prog=f"{python_executable} -m glyphsLib")
    subparsers = parser.add_subparsers()

    parser_glyphs2ufo = subparsers.add_parser("glyphs2ufo", help=glyphs2ufo.__doc__)
    parser_glyphs2ufo.set_defaults(func=glyphs2ufo)
    parser_glyphs2ufo.add_argument(
        "--version", action="version", version="glyphsLib %s" % glyphsLib.__version__
    )
    parser_glyphs2ufo.add_argument(
        "glyphs_file", metavar="GLYPHS_FILE", help="Glyphs file to convert."
    )
    parser_glyphs2ufo.add_argument(
        "-m",
        "--output-dir",
        default=None,
        help="Output directory of masters. (default: directory of Glyphs file)",
    )
    parser_glyphs2ufo.add_argument(
        "-d",
        "--designspace-path",
        default=None,
        help="Output path of designspace file. (default: directory of Glyphs file)",
    )
    parser_glyphs2ufo.add_argument(
        "-n",
        "--instance-dir",
        default=None,
        help=(
            "Output directory of instances. (default: output_dir/instance_ufos"
            "). This sets the file path for instances inside the designspace "
            "file."
        ),
    )
    parser_glyphs2ufo.add_argument(
        "--ufo-module",
        metavar="UFO_MODULE",
        choices=("ufoLib2", "defcon"),
        default="ufoLib2",
        help=(
            "Select the default library for writing UFOs. Choose between: %(choices)s "
            "(default: %(default)s)"
        ),
    )
    group = parser_glyphs2ufo.add_argument_group(
        "Roundtripping between Glyphs and UFOs"
    )
    group.add_argument(
        "--minimal",
        action="store_true",
        help=("Create minimal UFO files with only data needed for building fonts."),
    )
    group.add_argument(
        "--no-preserve-glyphsapp-metadata",
        action="store_false",
        help=(
            "Skip preserving Glyphs metadata in master UFOs and designspace "
            "file, which would be used to minimize differences when "
            "roundtripping between Glyphs and UFOs."
        ),
    )
    anchor_group = group.add_mutually_exclusive_group()
    anchor_group.add_argument(
        "--propagate-anchors",
        default=None,
        action="store_true",
        help=(
            "Copy anchors from underlying components to actual glyph "
            "(Glyphs would do this implicitly upon export). By default, if this option "
            "is not explicitly used, glyphs2ufo checks the 'Propagate Anchors' font "
            "custom parameter and if unset or set to 1 (true) then the anchors get "
            "propagated. Pass this option to enable this overriding the parameter."
        ),
    )
    anchor_group.add_argument(
        "--no-propagate-anchors",
        dest="propagate_anchors",
        default=None,
        action="store_false",
        help="Disable anchor propagation regardless of any custom parameter.",
    )
    gdef_gen_group = group.add_mutually_exclusive_group()
    gdef_gen_group.add_argument(
        "--generate-GDEF",
        action="store_true",
        default=False,
        dest="generate_GDEF",
        help=(
            "Write GDEF categories into the UFO's `public.openTypeCatgeories` lib key."
        ),
    )
    gdef_gen_group.add_argument(
        "--no-generate-GDEF",
        action="store_false",
        default=False,
        dest="generate_GDEF",
        help=(
            "Skip writing GDEF categories into the UFO's "
            "`public.openTypeCatgeories` lib key (default behavior, for backward "
            "compatibility)."
        ),
    )
    group.add_argument(
        "-N",
        "--normalize-ufos",
        action="store_true",
        help=(
            "Normalize UFOs with ufonormalizer, which avoids "
            "differences due to spacing, reordering of keys, etc."
        ),
    )
    group.add_argument(
        "--create-background-layers",
        action="store_true",
        help=(
            "Create background layers for all glyphs unless present, "
            "this can help in a workflow with multiple tools that "
            "may create background layers automatically."
        ),
    )
    group.add_argument(
        "--no-store-editor-state",
        action="store_true",
        help=(
            "Skip storing editor state in the UFO, like which glyphs are open "
            "in which tab (DisplayStrings)."
        ),
    )
    group.add_argument(
        "--write-public-skip-export-glyphs",
        action="store_true",
        help=(
            "Store the glyph export flag in the `public.skipExportGlyphs` list "
            "instead of the glyph-level 'com.schriftgestaltung.Glyphs.Export' lib "
            "key."
        ),
    )
    group.add_argument(
        "--expand-includes",
        action="store_true",
        help=(
            "Expand include statements in the .glyphs features and inline them in "
            "the exported UFO features.fea."
        ),
    )
    group = parser_glyphs2ufo.add_argument_group("Glyph data")
    group.add_argument(
        "--glyph-data",
        action="append",
        metavar="GLYPHDATA",
        help=(
            "Custom GlyphData XML file with glyph info (production name, "
            "script, category, subCategory, etc.). Can be used more than once."
        ),
    )

    parser_ufo2glyphs = subparsers.add_parser("ufo2glyphs", help=ufo2glyphs.__doc__)
    parser_ufo2glyphs.set_defaults(func=ufo2glyphs)
    parser_ufo2glyphs.add_argument(
        "--version", action="version", version="glyphsLib %s" % glyphsLib.__version__
    )
    parser_ufo2glyphs.add_argument(
        "designspace_file_or_UFOs",
        nargs="+",
        metavar="DESIGNSPACE_FILE_OR_UFOS",
        help=(
            "A single designspace file *or* one or more UFOs to convert to "
            "a Glyphs file."
        ),
    )
    parser_ufo2glyphs.add_argument(
        "--output-path", help="The path to write the Glyphs file to."
    )
    parser_ufo2glyphs.add_argument(
        "--ufo-module",
        metavar="UFO_MODULE",
        choices=("ufoLib2", "defcon"),
        default="ufoLib2",
        help=(
            "Select the default library for reading UFOs. Choose between: %(choices)s "
            "(default: %(default)s)"
        ),
    )
    group = parser_ufo2glyphs.add_argument_group(
        "Roundtripping between UFOs and Glyphs"
    )
    group.add_argument(
        "--no-preserve-glyphsapp-metadata",
        action="store_false",
        help=(
            "Skip preserving Glyphs metadata in master UFOs and designspace "
            "file, which would be used to minimize differences when "
            "roundtripping between Glyphs and UFOs."
        ),
    )
    group.add_argument(
        "--enable-last-change",
        action="store_false",
        help="Store modification timestamp in glyphs. Unnecessary when using Git.",
    )
    group.add_argument(
        "--enable-automatic-alignment",
        action="store_false",
        help="Enable automatic alignment of components in glyphs.",
    )
    group.add_argument(
        "--expand-includes",
        action="store_true",
        help=(
            "Expand include statements in the UFO features.fea and inline them in "
            "the exported .glyphs features."
        ),
    )

    options = parser.parse_args(args)

    if "func" in vars(options):
        return options.func(options)
    else:
        parser.print_help()


def glyphs2ufo(options):
    """Converts a Glyphs.app source file into UFO masters and a designspace file."""
    if options.output_dir is None:
        options.output_dir = os.path.dirname(options.glyphs_file) or "."

    if options.designspace_path is None:
        options.designspace_path = os.path.join(
            options.output_dir,
            os.path.basename(os.path.splitext(options.glyphs_file)[0]) + ".designspace",
        )

    # If options.instance_dir is None, instance UFO paths in the designspace
    # file will either use the value in customParameter's UFO_FILENAME_CUSTOM_PARAM or
    # be made relative to "instance_ufos/".
    glyphsLib.build_masters(
        options.glyphs_file,
        options.output_dir,
        options.instance_dir,
        designspace_path=options.designspace_path,
        minimize_glyphs_diffs=options.no_preserve_glyphsapp_metadata,
        propagate_anchors=options.propagate_anchors,
        normalize_ufos=options.normalize_ufos,
        create_background_layers=options.create_background_layers,
        generate_GDEF=options.generate_GDEF,
        store_editor_state=not options.no_store_editor_state,
        write_skipexportglyphs=options.write_public_skip_export_glyphs,
        expand_includes=options.expand_includes,
        ufo_module=__import__(options.ufo_module),
        minimal=options.minimal,
        glyph_data=options.glyph_data or None,
    )


def _glyphs2ufo_entry_point():
    """Provides entry point for a script to keep argparsing in main()."""
    args = sys.argv[1:]
    args.insert(0, "glyphs2ufo")
    return main(args)


def ufo2glyphs(options):
    """Convert one designspace file or one or more UFOs to a Glyphs.app source file."""
    import fontTools.designspaceLib
    from glyphsLib.util import open_ufo

    ufo_module = __import__(options.ufo_module)

    sources = options.designspace_file_or_UFOs
    designspace_file = None
    if (
        len(sources) == 1
        and sources[0].endswith(".designspace")
        and os.path.isfile(sources[0])
    ):
        designspace_file = sources[0]
        designspace = fontTools.designspaceLib.DesignSpaceDocument()
        designspace.read(designspace_file)
        object_to_read = designspace
    elif all(source.endswith(".ufo") and os.path.isdir(source) for source in sources):
        ufos = [open_ufo(source, ufo_module.Font) for source in sources]
        ufos.sort(
            key=lambda ufo: [  # Order the masters by weight and width
                ufo.info.openTypeOS2WeightClass or 400,
                ufo.info.openTypeOS2WidthClass or 5,
            ]
        )
        object_to_read = ufos
    else:
        print(
            "Please specify just one designspace file *or* one or more "
            "UFOs. They must end in '.designspace' or '.ufo', respectively.",
            file=sys.stderr,
        )
        return 1

    font = glyphsLib.to_glyphs(
        object_to_read,
        ufo_module=ufo_module,
        minimize_ufo_diffs=options.no_preserve_glyphsapp_metadata,
        expand_includes=options.expand_includes,
    )

    # Make the Glyphs file more suitable for roundtrip:
    font.customParameters["Disable Last Change"] = options.enable_last_change
    font.disablesAutomaticAlignment = options.enable_automatic_alignment

    if options.output_path:
        font.save(options.output_path)
    else:
        if designspace_file:
            filename_to_write = os.path.splitext(designspace_file)[0] + ".glyphs"
        else:
            filename_to_write = os.path.join(
                os.path.dirname(sources[0]),
                font.familyName.replace(" ", "") + ".glyphs",
            )
        font.save(filename_to_write)


def _ufo2glyphs_entry_point():
    """Provides entry point for a script to keep argparsing in main()."""
    args = sys.argv[1:]
    args.insert(0, "ufo2glyphs")
    return main(args)



#
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os
import fnmatch
from fontTools.designspaceLib import DesignSpaceDocument
from xmldiff import main, formatting
from fontTools.varLib import FEAVAR_FEATURETAG_LIB_KEY

import itertools
import pytest

import glyphsLib
from glyphsLib import to_designspace, to_glyphs
from glyphsLib.builder.constants import CUSTOM_PARAMETERS_KEY
from glyphsLib.util import open_ufo
from glyphsLib.types import Point


def test_designspace_generation_regular_same_family_name(tmpdir, ufo_module):
    ufo_Lt = ufo_module.Font()
    ufo_Lt.info.familyName = "CoolFoundry Examplary Serif"
    ufo_Lt.info.styleName = "Light"
    ufo_Lt.info.openTypeOS2WeightClass = 300

    ufo_Rg = ufo_module.Font()
    ufo_Rg.info.familyName = "CoolFoundry Examplary Serif"
    ufo_Rg.info.styleName = "Regular"
    ufo_Rg.info.openTypeOS2WeightClass = 400

    ufo_Md = ufo_module.Font()
    ufo_Md.info.familyName = "CoolFoundry Examplary Serif"
    ufo_Md.info.styleName = "Medium"
    ufo_Md.info.openTypeOS2WeightClass = 500

    ufo_Bd = ufo_module.Font()
    ufo_Bd.info.familyName = "CoolFoundry Examplary Serif"
    ufo_Bd.info.styleName = "Bold"
    ufo_Bd.info.openTypeOS2WeightClass = 700

    ufo_ExBd = ufo_module.Font()
    ufo_ExBd.info.familyName = "CoolFoundry Examplary Serif"
    ufo_ExBd.info.styleName = "XBold"
    ufo_ExBd.info.openTypeOS2WeightClass = 800

    font = to_glyphs([ufo_Lt, ufo_Rg, ufo_Md, ufo_Bd, ufo_ExBd])
    designspace = to_designspace(font, ufo_module=ufo_module)

    path = os.path.join(str(tmpdir), "actual.designspace")
    designspace.write(path)

    expected_path = os.path.join(
        os.path.dirname(__file__), "..", "data", "DesignspaceGenTestRegular.designspace"
    )

    assert (
        len(main.diff_files(path, expected_path, formatter=formatting.DiffFormatter()))
        == 0
    )


def test_designspace_generation_italic_same_family_name(tmpdir, ufo_module):
    ufo_Lt = ufo_module.Font()
    ufo_Lt.info.familyName = "CoolFoundry Examplary Serif"
    ufo_Lt.info.styleName = "Light Italic"
    ufo_Lt.info.openTypeOS2WeightClass = 300
    ufo_Lt.info.italicAngle = -11

    ufo_Rg = ufo_module.Font()
    ufo_Rg.info.familyName = "CoolFoundry Examplary Serif"
    ufo_Rg.info.styleName = "Regular Italic"
    ufo_Rg.info.openTypeOS2WeightClass = 400
    ufo_Rg.info.italicAngle = -11

    ufo_Md = ufo_module.Font()
    ufo_Md.info.familyName = "CoolFoundry Examplary Serif"
    ufo_Md.info.styleName = "Medium Italic"
    ufo_Md.info.openTypeOS2WeightClass = 500
    ufo_Md.info.italicAngle = -11

    ufo_Bd = ufo_module.Font()
    ufo_Bd.info.familyName = "CoolFoundry Examplary Serif"
    ufo_Bd.info.styleName = "Bold Italic"
    ufo_Bd.info.openTypeOS2WeightClass = 700
    ufo_Bd.info.italicAngle = -11

    ufo_ExBd = ufo_module.Font()
    ufo_ExBd.info.familyName = "CoolFoundry Examplary Serif"
    ufo_ExBd.info.styleName = "XBold Italic"
    ufo_ExBd.info.openTypeOS2WeightClass = 800
    ufo_ExBd.info.italicAngle = -11

    font = to_glyphs([ufo_Lt, ufo_Rg, ufo_Md, ufo_Bd, ufo_ExBd])
    designspace = to_designspace(font, ufo_module=ufo_module)

    path = os.path.join(str(tmpdir), "actual.designspace")
    designspace.write(path)

    expected_path = os.path.join(
        os.path.dirname(__file__), "..", "data", "DesignspaceGenTestItalic.designspace"
    )

    assert (
        len(main.diff_files(path, expected_path, formatter=formatting.DiffFormatter()))
        == 0
    )


def test_designspace_generation_regular_different_family_names(tmpdir, ufo_module):
    ufo_Lt = ufo_module.Font()
    ufo_Lt.info.familyName = "CoolFoundry Examplary Serif Light"
    ufo_Lt.info.styleName = "Regular"
    ufo_Lt.info.openTypeOS2WeightClass = 300

    ufo_Rg = ufo_module.Font()
    ufo_Rg.info.familyName = "CoolFoundry Examplary Serif"
    ufo_Rg.info.styleName = "Regular"
    ufo_Rg.info.openTypeOS2WeightClass = 400

    # Different family names are not allowed
    # REVIEW: reasonable requirement?
    with pytest.raises(Exception):
        to_glyphs([ufo_Lt, ufo_Rg])


def test_designspace_generation_same_weight_name(tmpdir, ufo_module):
    ufo_Bd = ufo_module.Font()
    ufo_Bd.info.familyName = "Test"
    ufo_Bd.info.styleName = "Bold"

    ufo_ExBd = ufo_module.Font()
    ufo_ExBd.info.familyName = "Test"
    ufo_ExBd.info.styleName = "Bold"

    ufo_XExBd = ufo_module.Font()
    ufo_XExBd.info.familyName = "Test"
    ufo_XExBd.info.styleName = "Bold"

    font = to_glyphs([ufo_Bd, ufo_ExBd, ufo_XExBd])
    designspace = to_designspace(font, ufo_module=ufo_module)

    assert designspace.sources[0].filename != designspace.sources[1].filename
    assert designspace.sources[1].filename != designspace.sources[2].filename
    assert designspace.sources[0].filename != designspace.sources[2].filename


@pytest.mark.parametrize("filename", ["BraceTestFont.glyphs", "BraceTestFontV3.glyphs"])
def test_designspace_generation_brace_layers(datadir, filename, ufo_module):
    with open(str(datadir.join(filename))) as f:
        font = glyphsLib.load(f)
    designspace = to_designspace(font, ufo_module=ufo_module, minimal=True)

    axes_order = [
        (a.name, a.minimum, a.default, a.maximum, a.map) for a in designspace.axes
    ]
    assert axes_order == [
        ("Width", 75, 100, 100, [(75, 50.0), (100, 100.0)]),
        ("Weight", 100, 100, 700, [(100, 100.0), (700, 1000.0)]),
    ]

    for (fname, layerName, name, ufo), (exp_fname, exp_layerName, exp_name) in zip(
        [(s.filename, s.layerName, s.name, s.font) for s in designspace.sources],
        [
            ("NewFont-Light.ufo", None, "New Font Light"),
            ("NewFont-Light.ufo", "{75}", "New Font Light {75}"),
            ("NewFont-Bold.ufo", None, "New Font Bold"),
            ("NewFont-Bold.ufo", "{75}", "New Font Bold {75}"),
            ("NewFont-Bold.ufo", "*{90.5, 500}", "New Font Bold *{90.5, 500}"),
            ("NewFont-Bold.ufo", "*{90.5, 600}", "New Font Bold *{90.5, 600}"),
            ("NewFont-CondensedLight.ufo", None, "New Font Condensed Light"),
            ("NewFont-CondensedBold.ufo", None, "New Font Condensed Bold"),
        ],
    ):
        assert fname == exp_fname
        # the brace layer name for Glyphs3 is automatically generated from its
        # coordinates and unlike in Glyphs2 won't contain any extra prefix, hence
        # we only 'fnmatch' it.
        # https://github.com/googlefonts/glyphsLib/issues/851
        if exp_layerName is not None:
            assert fnmatch.fnmatch(layerName, exp_layerName)
            assert layerName in ufo.layers
        else:
            assert layerName is None
        assert fnmatch.fnmatch(name, exp_name)

    # Check that all sources have a font object attached and sources with the same
    # filename have the same font object attached.
    masters = {}
    for source in designspace.sources:
        assert source.font
        if source.filename in masters:
            assert masters[source.filename] is source.font
        masters[source.filename] = source.font

    # Check that brace layer glyph is created
    assert len(designspace.sources[0].font.layers) == 2


# https://github.com/googlefonts/glyphsLib/issues/1097
def test_designspace_generation_v3_brace_component_without_a_name(datadir, ufo_module):
    with open(str(datadir.join("BraceCompositev3.glyphs"))) as f:
        font = glyphsLib.load(f)
    designspace = to_designspace(font, ufo_module=ufo_module)
    light = designspace.sources[0]
    assert [l.name for l in light.font.layers] == ["public.default", "{100}"]
    assert "aacute" in light.font.layers["{100}"]


@pytest.mark.parametrize("filename", ["BraceTestFont.glyphs", "BraceTestFontV3.glyphs"])
def test_designspace_generation_instances(datadir, filename, ufo_module):
    with open(str(datadir.join(filename))) as f:
        font = glyphsLib.load(f)
    designspace = to_designspace(font, ufo_module=ufo_module)

    instances_order = [
        (i.name, i.styleMapStyleName, i.location) for i in designspace.instances
    ]
    assert instances_order == [
        ("New Font Thin", "regular", {"Width": 100.0, "Weight": 100.0}),
        ("New Font Regular", "regular", {"Width": 100.0, "Weight": 500.0}),
        ("New Font Bold", "bold", {"Width": 100.0, "Weight": 1000.0}),
        ("New Font Semi Consensed", "regular", {"Width": 75.0, "Weight": 500.0}),
        ("New Font Thin Condensed", "regular", {"Width": 50.0, "Weight": 100.0}),
        ("New Font Condensed", "regular", {"Width": 50.0, "Weight": 500.0}),
        ("New Font Bold Condensed", "bold", {"Width": 50.0, "Weight": 1000.0}),
    ]


@pytest.mark.parametrize("filename", ["BraceTestFont.glyphs", "BraceTestFontV3.glyphs"])
def test_designspace_generation_on_disk(datadir, tmpdir, filename, ufo_module):
    glyphsLib.build_masters(str(datadir.join(filename)), str(tmpdir))

    ufo_paths = list(tmpdir.visit(fil="*.ufo"))
    assert len(ufo_paths) == 4  # Source layers should not be written to disk.
    for ufo_path in ufo_paths:
        ufo = open_ufo(ufo_path, ufo_module.Font)

        # Check that all glyphs have contours (brace layers are in "b" only, writing
        # the brace layer to disk would result in empty other glyphs).
        for layer in ufo.layers:
            for glyph in layer:
                if glyph.name == "space":
                    assert not glyph
                else:
                    assert glyph


def test_designspace_generation_bracket_roundtrip(datadir, ufo_module):
    with open(str(datadir.join("BracketTestFont.glyphs"))) as f:
        font = glyphsLib.load(f)
    designspace = to_designspace(font, ufo_module=ufo_module)

    assert designspace.rules[0].name == "BRACKET.Weight_600_1000"
    assert designspace.rules[0].conditionSets == [
        [dict(name="Weight", minimum=600, maximum=1000)]
    ]
    assert sorted(designspace.rules[0].subs) == [
        ("a", "a.BRACKET.varAlt01"),
        ("x", "x.BRACKET.varAlt02"),
    ]

    assert designspace.rules[1].name == "BRACKET.Weight_300_600"
    assert designspace.rules[1].conditionSets == [
        [dict(name="Weight", minimum=300, maximum=600)]
    ]
    assert sorted(designspace.rules[1].subs) == [
        ("a", "a.BRACKET.varAlt01"),
        ("x", "x.BRACKET.varAlt01"),
    ]

    for source in designspace.sources:
        assert "[300]" not in source.font.layers
        assert "Something [300]" not in source.font.layers
        assert "[600]" not in source.font.layers
        assert "Other [600]" not in source.font.layers
        g1 = source.font["x.BRACKET.varAlt01"]
        assert not g1.unicodes
        g2 = source.font["x.BRACKET.varAlt02"]
        assert not g2.unicodes

    font_rt = to_glyphs(designspace)
    assert "x" in font_rt.glyphs
    g1 = font_rt.glyphs["x"]
    assert len(g1.layers) == 12 and {l.name for l in g1.layers} == {
        "[300]",
        "[600]",
        "Bold",
        "Condensed Bold",
        "Condensed Light",
        "Light",
        "Other [600]",
        "Something [300]",
    }
    g2 = font_rt.glyphs["a"]
    assert len(g2.layers) == 8 and {l.name for l in g2.layers} == {
        "[300]",
        "Bold",
        "Condensed Bold",
        "Condensed Light",
        "Light",
    }
    assert "a.BRACKET.300" not in font_rt.glyphs
    assert "x.BRACKET.300" not in font_rt.glyphs
    assert "x.BRACKET.600" not in font_rt.glyphs


def test_designspace_generation_bracket_roundtrip_v3(datadir, ufo_module):
    # v3 supports bracket layers that reference positions on multiple axes.
    # of particular interest for this test is that we don't bother including
    # irrelevant axes in the generated rules; that is, if a rule matches across
    # all of an axis, it should be omitted.
    with open(str(datadir.join("BracketTestFont_v3.glyphs"))) as f:
        font = glyphsLib.load(f)
    designspace = to_designspace(font, ufo_module=ufo_module)

    assert designspace.rules[0].name == "BRACKET.Weight_600_1000.Width_50_80"
    assert designspace.rules[0].conditionSets == [
        [
            dict(name="Weight", minimum=600, maximum=1000),
            dict(name="Width", minimum=50, maximum=80),
        ]
    ]
    assert sorted(designspace.rules[0].subs) == [
        ("a", "a.BRACKET.varAlt01"),
        ("b", "b.BRACKET.varAlt01"),
        ("x", "x.BRACKET.varAlt02"),
    ]

    assert designspace.rules[1].name == "BRACKET.Weight_600_1000"
    assert designspace.rules[1].conditionSets == [
        [dict(name="Weight", minimum=600, maximum=1000)]
    ]
    assert sorted(designspace.rules[1].subs) == [
        ("a", "a.BRACKET.varAlt01"),
        ("x", "x.BRACKET.varAlt02"),
    ]

    assert designspace.rules[2].name == "BRACKET.Weight_300_600.Width_50_80"
    assert designspace.rules[2].conditionSets == [
        [
            dict(name="Weight", minimum=300, maximum=600),
            dict(name="Width", minimum=50, maximum=80),
        ]
    ]
    assert sorted(designspace.rules[2].subs) == [
        ("a", "a.BRACKET.varAlt01"),
        ("b", "b.BRACKET.varAlt01"),
        ("x", "x.BRACKET.varAlt01"),
    ]

    assert designspace.rules[3].name == "BRACKET.Weight_300_600"
    assert designspace.rules[3].conditionSets == [
        [dict(name="Weight", minimum=300, maximum=600)]
    ]
    assert sorted(designspace.rules[3].subs) == [
        ("a", "a.BRACKET.varAlt01"),
        ("x", "x.BRACKET.varAlt01"),
    ]

    for source in designspace.sources:
        assert "[300]" not in source.font.layers
        assert "Something [300]" not in source.font.layers
        assert "[600]" not in source.font.layers
        assert "Other [600]" not in source.font.layers
        g1 = source.font["x.BRACKET.varAlt01"]
        assert not g1.unicodes
        g2 = source.font["x.BRACKET.varAlt02"]
        assert not g2.unicodes

    font_rt = to_glyphs(designspace)
    assert "x" in font_rt.glyphs
    g1 = font_rt.glyphs["x"]
    assert len(g1.layers) == 12 and {l.name for l in g1.layers} == {
        "[300]",
        "[600]",
        "Bold",
        "Bold Condensed",
        "Light Condensed",
        "Light",
        "Other [600]",
        "Something [300]",
    }
    g2 = font_rt.glyphs["a"]
    assert len(g2.layers) == 8 and {l.name for l in g2.layers} == {
        "[300]",
        "Bold",
        "Bold Condensed",
        "Light Condensed",
        "Light",
    }
    assert "a.BRACKET.300" not in font_rt.glyphs
    assert "x.BRACKET.300" not in font_rt.glyphs
    assert "x.BRACKET.600" not in font_rt.glyphs


def test_designspace_generation_bracket_roundtrip_psnames(datadir, ufo_module):
    with open(str(datadir.join("PSNames.glyphs"))) as f:
        font = glyphsLib.load(f)
    designspace: DesignSpaceDocument = to_designspace(font, ufo_module=ufo_module)

    assert designspace.findDefault().font.lib["public.postscriptNames"] == {
        "a-cy": "uni0430",
        "a-cy.BRACKET.varAlt01": "uni0430.BRACKET.varAlt01",
        "a-cy.alt": "uni0430.alt",
    }

    font_rt = to_glyphs(designspace)
    designspace_rt = to_designspace(font_rt, ufo_module=ufo_module)

    assert designspace_rt.findDefault().font.lib["public.postscriptNames"] == {
        "a-cy": "uni0430",
        "a-cy.BRACKET.varAlt01": "uni0430.BRACKET.varAlt01",
        "a-cy.alt": "uni0430.alt",
    }

    font_rt2 = to_glyphs(designspace_rt)
    designspace_rt2 = to_designspace(font_rt2, ufo_module=ufo_module)

    assert designspace_rt2.findDefault().font.lib["public.postscriptNames"] == {
        "a-cy": "uni0430",
        "a-cy.BRACKET.varAlt01": "uni0430.BRACKET.varAlt01",
        "a-cy.alt": "uni0430.alt",
    }


@pytest.mark.xfail
def test_designspace_generation_bracket_roundtrip_no_layername(datadir, ufo_module):
    with open(str(datadir.join("BracketTestFont.glyphs"))) as f:
        font = glyphsLib.load(f)

    # Remove brace layers for clean slate.
    master_ids = {m.id for m in font.masters}
    for g in font.glyphs:
        dl = [l for l in g.layers if l.layerId not in master_ids]
        for l in dl:
            g.layers.remove(l)

    designspace = to_designspace(font, ufo_module=ufo_module)
    for source in designspace.sources:
        source.font.newGlyph("b.BRACKET.100")

    font_rt = to_glyphs(designspace)
    for layer in font_rt.glyphs["b"].layers:
        if layer.layerId not in master_ids:
            assert layer.name == "[100]"


def test_designspace_generation_bracket_unbalanced_brackets(datadir, ufo_module):
    with open(str(datadir.join("BracketTestFont2.glyphs"))) as f:
        font = glyphsLib.load(f)

    layer_names = {l.name for l in font.glyphs["C"].layers}
    assert layer_names == {"Regular", "Bold", "Bold [600]"}

    designspace = to_designspace(font, ufo_module=ufo_module)

    for source in designspace.sources:
        assert "C.BRACKET.varAlt01" in source.font

    font_rt = to_glyphs(designspace)

    assert "C" in font_rt.glyphs

    assert {l.name for l in font_rt.glyphs["C"].layers} == layer_names
    assert "C.BRACKET.varAlt01" not in font_rt.glyphs


def test_designspace_generation_bracket_composite_glyph(datadir, ufo_module):
    with open(str(datadir.join("BracketTestFont2.glyphs"))) as f:
        font = glyphsLib.load(f)

    g = font.glyphs["B"]
    for layer in g.layers:
        assert layer.components[0].name == "A"

    designspace = to_designspace(font, ufo_module=ufo_module)

    for source in designspace.sources:
        ufo = source.font
        assert "B.BRACKET.varAlt01" in ufo
        assert ufo["B"].components[0].baseGlyph == "A"
        assert ufo["B.BRACKET.varAlt01"].components[0].baseGlyph == "A.BRACKET.varAlt01"
        # G has no alternate layers, but it uses a component
        # which does, so it too must develop some.
        assert "G.BRACKET.varAlt01" in ufo

    font_rt = to_glyphs(designspace)

    assert "B" in font_rt.glyphs

    g2 = font_rt.glyphs["B"]
    for layer in g2.layers:
        assert layer.components[0].name == "A"

    assert "B.BRACKET.600" not in font_rt.glyphs


def test_designspace_generation_reverse_bracket_roundtrip(datadir, ufo_module):
    with open(str(datadir.join("BracketTestFont2.glyphs"))) as f:
        font = glyphsLib.load(f)

    g = font.glyphs["D"]

    assert {"Regular ]600]", "Bold ]600]"}.intersection(l.name for l in g.layers)

    designspace = to_designspace(font, ufo_module=ufo_module)

    # Bottom box should include substitutions for D (400->600)
    assert designspace.rules[2].name == "BRACKET.Weight_400_570"
    assert designspace.rules[2].conditionSets == [
        [dict(name="Weight", minimum=400, maximum=570)]
    ]
    assert designspace.rules[2].subs == [
        ("D", "D.BRACKET.varAlt01"),
        ("E", "E.BRACKET.varAlt01"),
        ("F", "F.BRACKET.varAlt02"),
    ]

    for source in designspace.sources:
        ufo = source.font
        assert "D.BRACKET.varAlt01" in ufo

    font_rt = to_glyphs(designspace)

    assert "D" in font_rt.glyphs

    g2 = font_rt.glyphs["D"]
    assert {"Regular ]600]", "Bold ]600]"}.intersection(l.name for l in g2.layers)

    assert "D.BRACKET.wght_400_600" not in font_rt.glyphs


def test_designspace_generation_bracket_no_export_glyph(datadir, ufo_module):
    with open(str(datadir.join("BracketTestFont2.glyphs"))) as f:
        font = glyphsLib.load(f)

    font.glyphs["E"].export = False

    designspace = to_designspace(
        font, write_skipexportglyphs=True, ufo_module=ufo_module, minimal=False
    )

    assert "E" in designspace.lib.get("public.skipExportGlyphs")

    for source in designspace.sources:
        assert "E.REV_BRACKET.570" not in source.font
        assert "E.BRACKET.630" not in source.font

    for rule in designspace.rules:
        assert "E" not in {g for g in itertools.chain(*rule.subs)}

    font_rt = to_glyphs(designspace)

    assert "E" in font_rt.glyphs
    assert {l.name for l in font_rt.glyphs["E"].layers} == {
        "Regular",
        "Regular [630]",
        "Bold",
        "Bold ]570]",
    }


def test_designspace_generation_bracket_GDEF(datadir, ufo_module):
    with open(str(datadir.join("BracketTestFont.glyphs"))) as f:
        font = glyphsLib.load(f)

    # add some attaching anchors to the "x" glyph and its (bracket) layers to
    # trigger the generation of GDEF table
    for layer in font.glyphs["x"].layers:
        anchor = glyphsLib.classes.GSAnchor()
        anchor.name = "top"
        anchor.position = Point(0, 0)
        layer.anchors.append(anchor)

    designspace = to_designspace(font, ufo_module=ufo_module, generate_GDEF=True)

    for source in designspace.sources:
        categories = source.font.lib["public.openTypeCategories"]

        assert categories == {
            "x": "base",
            "x.BRACKET.varAlt01": "base",
            "x.BRACKET.varAlt02": "base",
        }


def test_designspace_generation_bracket_glyphs3_simple(datadir, ufo_module):
    with open(str(datadir.join("Alternate-g3-axis1.glyphs"))) as f:
        font = glyphsLib.load(f)

    designspace = to_designspace(font, ufo_module=ufo_module)

    for source in designspace.sources:
        assert "A.BRACKET.varAlt01" in source.font


def test_designspace_generation_bracket_rclt_roundtrip(datadir, ufo_module):
    with open(str(datadir.join("BracketTestFont.glyphs"))) as f:
        font = glyphsLib.load(f)
    font.customParameters["Feature for Feature Variations"] = "rclt"
    designspace = to_designspace(font, ufo_module=ufo_module)
    assert designspace.rulesProcessingLast

    font_rt = to_glyphs(designspace)
    assert font_rt.customParameters["Feature for Feature Variations"] == "rclt"


def test_designspace_generation_bracket_other_roundtrip(datadir, ufo_module):
    with open(str(datadir.join("BracketTestFont.glyphs"))) as f:
        font = glyphsLib.load(f)
    font.customParameters["Feature for Feature Variations"] = "calt"
    designspace = to_designspace(font, ufo_module=ufo_module)
    assert FEAVAR_FEATURETAG_LIB_KEY in designspace.lib
    assert designspace.lib[FEAVAR_FEATURETAG_LIB_KEY] == "calt"

    font_rt = to_glyphs(designspace)
    assert font_rt.customParameters["Feature for Feature Variations"] == "calt"


def test_designspace_generation_multiaxis_bracket(datadir, ufo_module):
    with open(str(datadir.join("Playfair-v.glyphs")), encoding="utf-8") as f:
        font = glyphsLib.load(f)

    # Remove names of bracket layers, make sure the layers get
    # copied anyway
    for l in font.glyphs["v"].layers:
        if l._is_bracket_layer():
            l.name = ""

    designspace = to_designspace(font, ufo_module=ufo_module)
    axes = designspace.axes
    info = font.glyphs["v"].layers[8]._bracket_info(axes)
    assert info == {"opsz": (5, 410), "wdth": (50, 75), "wght": (690, 900)}

    for source in designspace.sources:
        assert "v.BRACKET.varAlt01" in source.font

    assert (
        designspace.rules[0].name
        == "BRACKET.Optical size_5_410.Weight_690_900.Width_50_75"
    )
    assert designspace.rules[0].conditionSets == [
        [
            dict(name="Optical size", minimum=5, maximum=410),
            dict(name="Width", minimum=50, maximum=75),
            dict(name="Weight", minimum=690, maximum=900),
        ]
    ]
    assert designspace.rules[0].subs == [
        ("v", "v.BRACKET.varAlt01"),
    ]


@pytest.mark.parametrize("minimal", [True, False])
def test_designspace_generation_instances_disabled_custom_params(
    datadir, ufo_module, minimal
):
    with open(str(datadir.join("GlyphsUnitTestSans.glyphs"))) as f:
        font = glyphsLib.load(f)

    designspace = to_designspace(font, ufo_module=ufo_module, minimal=minimal)

    regular_instance = next(
        (i for i in designspace.instances if i.name.endswith("Regular"))
    )
    if minimal:
        assert len(regular_instance.lib) == 0
    else:
        assert len(regular_instance.lib) == 1
        assert CUSTOM_PARAMETERS_KEY in regular_instance.lib
        assert regular_instance.lib[CUSTOM_PARAMETERS_KEY] == [("fsType", [])]



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from collections import defaultdict
import os

from glyphsLib import classes
from .constants import (
    UFO_ORIGINAL_KERNING_GROUPS_KEY,
    UFO_GROUPS_NOT_IN_FEATURE_KEY,
    UFO_KERN_GROUP_PATTERN,
    BRACKET_GLYPH_RE,
)


def _get_glyphs_with_rtl_kerning(font):
    # Return a set of all glyph names that are referenced from font.kerningRTL,
    # either directly as single glyphs or as part of kerning groups.

    rtl_glyphs = set()
    if not font.kerningRTL:
        return rtl_glyphs

    # sets of group names keyed by {left,right}KerningGroup
    rtl_groups = defaultdict(set)
    glyph_kerning_attr = {"R": "leftKerningGroup", "L": "rightKerningGroup"}

    def mark_as_rtl(s, side):
        if s.startswith(f"@MMK_{side}_"):
            rtl_groups[glyph_kerning_attr[side]].add(s[7:])
        else:  # single glyph
            assert not s.startswith("@MMK_"), f"unexpected key in kerningRTL: {s}"
            rtl_glyphs.add(s)

    for kerning_id in {
        m.id if m.metricsSource is None else m.metricsSource.id for m in font.masters
    }:
        for kern1, subtable in font.kerningRTL.get(kerning_id, {}).items():
            mark_as_rtl(kern1, side="R")
            for kern2 in subtable.keys():
                mark_as_rtl(kern2, side="L")

    for glyph in font.glyphs.values():
        if glyph.name not in rtl_glyphs and any(
            getattr(glyph, attr) in rtl_groups[attr]
            for attr in glyph_kerning_attr.values()
        ):
            rtl_glyphs.add(glyph.name)

    return rtl_glyphs


def to_ufo_groups(self):
    # Build groups once and then apply to all UFOs.
    groups = defaultdict(list)

    # Classes usually go to the feature file, unless we have our custom flag
    group_names = None
    if UFO_GROUPS_NOT_IN_FEATURE_KEY in self.font.userData.keys():
        group_names = set(self.font.userData[UFO_GROUPS_NOT_IN_FEATURE_KEY])
    if group_names:
        for gsclass in self.font.classes.values():
            if gsclass.name in group_names:
                if gsclass.code:
                    groups[gsclass.name] = gsclass.code.split(" ")
                else:
                    # Empty group: using split like above would produce ['']
                    groups[gsclass.name] = []

    # Rebuild kerning groups from `left/rightKerningGroup`s
    # Use the original list of kerning groups as a base, to recover
    #  - the original ordering
    #  - the kerning groups of glyphs that were not in the font (which can be
    #    stored in a UFO but not by Glyphs)
    recovered = set()
    orig_groups = self.font.userData.get(UFO_ORIGINAL_KERNING_GROUPS_KEY)
    if orig_groups:
        for group, glyphs in orig_groups.items():
            if not glyphs:
                # Restore empty group
                groups[group] = []
            for glyph_name in glyphs:
                # Check that the original value is still valid
                match = UFO_KERN_GROUP_PATTERN.match(group)
                side = match.group(1)
                group_name = match.group(2)
                glyph = self.font.glyphs[glyph_name]
                if not glyph or getattr(glyph, _glyph_kerning_attr(side)) == group_name:
                    # The original grouping is still valid
                    groups[group].append(glyph_name)
                    # Remember not to add this glyph again later
                    # Thus the original position in the list is preserved
                    recovered.add((glyph_name, int(side)))

    # Read new/modified grouping values.
    # For glyphs that are used in Glyphs3's kerningRTL dict, take the opposite side:
    # NOTE: Not only this breaks Glyphs<=>UFO round-tripping, but also it makes
    # impossible for the same glyph to be kerned in both LTR and RTL dictionaries.
    # While this is unfortunate, we believe it's better than completely ignoring
    # all Glyphs3's RTL kerning.
    # For more info: https://github.com/googlefonts/glyphsLib/pull/778
    rtl_glyphs = _get_glyphs_with_rtl_kerning(self.font)
    for glyph in self.font.glyphs.values():
        is_rtl = glyph.name in rtl_glyphs
        for side in 1, 2:
            if (glyph.name, side) not in recovered:
                attr = _glyph_kerning_attr(side, is_rtl)
                group = getattr(glyph, attr)
                if group:
                    group = f"public.kern{side}.{group}"
                    groups[group].append(glyph.name)

    # Update all UFOs with the same info
    for source in self._sources.values():
        for name, glyphs in groups.items():
            # Shallow copy to prevent unexpected object sharing
            source.font.groups[name] = glyphs[:]


def to_glyphs_groups(self):
    # Build the GSClasses from the groups of the first UFO.
    groups = []
    for source in self._sources.values():
        for name, glyphs in source.font.groups.items():
            # Filter out all BRACKET glyphs first, as they are created at
            # to_designspace time to inherit glyph kerning to their bracket
            # variants. They need to be removed because Glpyhs.app handles that
            # on its own.
            glyphs = [name for name in glyphs if not BRACKET_GLYPH_RE.match(name)]
            if _is_kerning_group(name):
                _to_glyphs_kerning_group(self, name, glyphs)
            else:
                gsclass = classes.GSClass(name, " ".join(glyphs))
                self.font.classes.append(gsclass)
                groups.append(name)
        if self.minimize_ufo_diffs:
            self.font.userData[UFO_GROUPS_NOT_IN_FEATURE_KEY] = groups
        break

    # Check that other UFOs are identical and print a warning if not.
    for index, source in enumerate(self._sources.values()):
        if index == 0:
            reference_ufo = source.font
        else:
            _assert_groups_are_identical(self, reference_ufo, source.font)


def _is_kerning_group(name):
    return name.startswith(("public.kern1.", "public.kern2."))


def _to_glyphs_kerning_group(self, name, glyphs):
    if self.minimize_ufo_diffs:
        # Preserve ordering when going from UFO group
        # to left/rightKerningGroup disseminated in GSGlyphs
        # back to UFO group.
        if not self.font.userData.get(UFO_ORIGINAL_KERNING_GROUPS_KEY):
            self.font.userData[UFO_ORIGINAL_KERNING_GROUPS_KEY] = {}
        self.font.userData[UFO_ORIGINAL_KERNING_GROUPS_KEY][name] = glyphs

    match = UFO_KERN_GROUP_PATTERN.match(name)
    side = match.group(1)
    group_name = match.group(2)
    for glyph_name in glyphs:
        glyph = self.font.glyphs[glyph_name]
        if glyph:
            setattr(glyph, _glyph_kerning_attr(side), group_name)


def _glyph_kerning_attr(side, is_rtl=False):
    """Return rightKerningGroup or leftKerningGroup depending on the UFO
    group's side (1 or 2).

    Flip values for RTL kerning.
    """
    side = int(side)
    assert side in (1, 2), f"invalid kerning side: {side}"
    if is_rtl:
        side = 2 if side == 1 else 1
    if side == 1:
        return "rightKerningGroup"
    else:
        return "leftKerningGroup"


def _assert_groups_are_identical(self, reference_ufo, ufo):
    first_time = [True]  # Using a mutable as a non-local for closure below

    def _warn(message, *args):
        if first_time:
            self.logger.warning(
                "Using UFO `%s` as a reference for groups:",
                _ufo_logging_ref(reference_ufo),
            )
            first_time.clear()
        self.logger.warning("   " + message, *args)

    # Check for inconsistencies
    for group, glyphs in ufo.groups.items():
        if group not in reference_ufo.groups:
            _warn(
                "group `%s` from `%s` will be lost because it's not "
                "defined in the reference UFO",
                group,
                _ufo_logging_ref(ufo),
            )
            continue
        reference_glyphs = reference_ufo.groups[group]
        if set(glyphs) != set(reference_glyphs):
            _warn(
                "group `%s` from `%s` will not be stored accurately because "
                "it is different from the reference UFO",
                group,
                _ufo_logging_ref(ufo),
            )
            _warn("    reference = %s", " ".join(sorted(reference_glyphs)))
            _warn("    current   = %s", " ".join(sorted(glyphs)))


def _ufo_logging_ref(ufo):
    """Return a string that can identify this UFO in logs."""
    if ufo.path:
        return os.path.basename(ufo.path)
    return ufo.info.styleName



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import argparse
import os

import glyphsLib


def glyphs_files(directory):
    for root, _dirs, files in os.walk(directory):
        for filename in files:
            if filename.endswith(".glyphs"):
                yield os.path.join(root, filename)


def main():
    parser = argparse.ArgumentParser(
        "Translate all .glyphs files into UFO+designspace in the specified directories."
    )
    parser.add_argument(
        "-o", "--out", metavar="OUTPUT_DIR", default="ufos", help="Output directory"
    )
    parser.add_argument("directories", nargs="*")
    args = parser.parse_args()

    for directory in args.directories:
        files = glyphs_files(directory)
        for filename in files:
            try:
                # Code for glyphsLib with roundtrip
                from glyphsLib.builder import to_designspace

                font = glyphsLib.GSFont(filename)
                designspace = to_designspace(font)
                dsname = font.familyName.replace(" ", "") + ".designspace"
                designspace.write(os.path.join(args.out, dsname))
            except ImportError:
                # This is the version that works with glyphsLib 2.1.0
                glyphsLib.build_masters(
                    filename, master_dir=args.out, designspace_instance_dir=args.out
                )


if __name__ == "__main__":
    main()



#
# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest
import unittest
import os

import glyphsLib
import glyphsLib.builder.constants
from glyphsLib import classes

from .. import test_helpers


class UFORoundtripTest(test_helpers.AssertUFORoundtrip):
    def test_empty_font(self):
        empty_font = classes.GSFont()
        empty_font.masters.append(classes.GSFontMaster())
        self.assertUFORoundtrip(empty_font)

    def test_GlyphsUnitTestSans(self):
        filename = os.path.join(
            os.path.dirname(__file__), "../data/GlyphsUnitTestSans.glyphs"
        )
        with open(filename) as f:
            font = glyphsLib.load(f)
        self.assertUFORoundtrip(font)

    @pytest.mark.xfail(reason="Master naming and instance data modification issues.")
    def test_BraceTestFont(self):
        filename = os.path.join(
            os.path.dirname(__file__), "../data/BraceTestFont.glyphs"
        )
        with open(filename) as f:
            font = glyphsLib.load(f)
        self.assertUFORoundtrip(font)

    def test_BraceTestFont_no_editor_state(self):
        filename = os.path.join(
            os.path.dirname(__file__), "../data/BraceTestFont.glyphs"
        )
        with open(filename) as f:
            font = glyphsLib.load(f)

        designspace = glyphsLib.to_designspace(font)
        for source in designspace.sources:
            assert source.font.lib[
                glyphsLib.builder.constants.FONT_CUSTOM_PARAM_PREFIX + "DisplayStrings"
            ] == ["a", "b"]

        designspace = glyphsLib.to_designspace(font, store_editor_state=False)
        for source in designspace.sources:
            assert (
                glyphsLib.builder.constants.FONT_CUSTOM_PARAM_PREFIX + "DisplayStrings"
                not in source.font.lib
            )


if __name__ == "__main__":
    unittest.main()



# Copyright 2016 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from .constants import HINTS_LIB_KEY
from glyphsLib.types import IndexPath
from glyphsLib.types import Point


def to_ufo_hints(self, ufo_glyph, layer):
    if not hasattr(layer, "hints"):
        return
    hints = []
    for source in layer.hints:
        hint = {}
        for name in ["horizontal", "options", "stem", "type", "name"]:
            hint[name] = getattr(source, name, None)
        for name in ["origin", "other1", "other2", "target"]:
            index_path = getattr(source, name, None)
            if index_path:
                if name == "target" and index_path.value in (["up"], ["down"]):
                    hint[name] = index_path.value[0]
                elif not any(value is None for value in index_path):
                    hint[name] = index_path.value
        for name in ["place", "scale"]:
            point = getattr(source, name, None)
            if point and not any(value is None for value in point):
                hint[name] = point.value
        hints.append(hint)
    if hints:
        ufo_glyph.lib[HINTS_LIB_KEY] = hints


def to_glyphs_hints(self, ufo_glyph, layer):
    if HINTS_LIB_KEY not in ufo_glyph.lib:
        return
    for source in ufo_glyph.lib[HINTS_LIB_KEY]:
        hint = self.glyphs_module.GSHint()
        for name in ["horizontal", "options", "stem", "type", "name"]:
            setattr(hint, name, source[name])
        for name in ["origin", "other1", "other2", "target"]:
            if name in source:
                value = source[name]
                # https://github.com/googlefonts/glyphsLib/pull/613
                # handle target = ['u', 'p'] or ['d', 'o', 'w', 'n']
                if name == "target" and value in ([list("down")], [list("up")]):
                    value = ["".join(value)]
                setattr(hint, name, IndexPath(*value))
        for name in ["place", "scale"]:
            if name in source:
                setattr(hint, name, Point(*source[name]))
        layer.hints.append(hint)



from glyphsLib.builder.filters import parse_glyphs_filter


def test_complete_parameter():
    inputstr = (
        "Transformations;LSB:+23;RSB:-22;SlantCorrection:true;"
        "OffsetX:10;OffsetY:-10;Origin:0;exclude:uni0334,uni0335 uni0336"
    )
    expected = {
        "name": "Transformations",
        "kwargs": {
            "LSB": 23,
            "RSB": -22,
            "SlantCorrection": True,
            "OffsetX": 10,
            "OffsetY": -10,
            "Origin": 0,
        },
        "exclude": ["uni0334", "uni0335", "uni0336"],
    }
    result = parse_glyphs_filter(inputstr)
    assert result == expected


def test_is_pre():
    inputstr = "Dummy"
    expected = {"name": "Dummy", "pre": True}
    result = parse_glyphs_filter(inputstr, is_pre=True)
    assert result == expected


def test_positional_parameter():
    inputstr = "Roughenizer;34;2;0;0.34"
    expected = {"name": "Roughenizer", "args": [34, 2, 0, 0.34]}
    result = parse_glyphs_filter(inputstr)
    assert result == expected


def test_single_name():
    inputstr = "AddExtremes"
    expected = {"name": "AddExtremes"}
    result = parse_glyphs_filter(inputstr)
    assert result == expected


def test_empty_string(caplog):
    inputstr = ""
    parse_glyphs_filter(inputstr)
    assert len(
        [r for r in caplog.records if "Failed to parse glyphs filter" in r.msg]
    ), "Empty string should trigger an error message"


def test_no_name(caplog):
    inputstr = ";OffsetX:2"
    parse_glyphs_filter(inputstr)
    assert len(
        [r for r in caplog.records if "Failed to parse glyphs filter" in r.msg]
    ), "Empty string with no filter name should trigger an error message"


def test_duplicate_exclude_include(caplog):
    inputstr = "thisisaname;34;-3.4;exclude:uni1111;include:uni0022;exclude:uni2222"
    expected = {"name": "thisisaname", "args": [34, -3.4], "exclude": ["uni2222"]}
    result = parse_glyphs_filter(inputstr)

    assert len(
        [r for r in caplog.records if "can only present as the last argument" in r.msg]
    ), (
        "The parse_glyphs_filter should warn user that the exclude/include "
        "should only be the last argument in the filter string."
    )
    assert result == expected


def test_empty_args_trailing_semicolon():
    inputstr = "thisisaname;3;;a:b;;;"
    expected = {"name": "thisisaname", "args": [3], "kwargs": {"a": "b"}}
    result = parse_glyphs_filter(inputstr)
    assert result == expected



import copy
import os

import glyphsLib

DATA = os.path.join(os.path.dirname(os.path.dirname(__file__)), "data")


def test_ufo_instance_properties():
    from glyphsLib.interpolation import apply_instance_data_to_ufo

    file = glyphsLib.GSFont(os.path.join(DATA, "UFOInstancePropertiesTest.glyphs"))
    for instance, name in zip(file.instances, ["Thin", "Regular", "Black"]):
        assert instance.preferredFamilyName == "Typographic Neue"
        assert instance.preferredSubfamilyName == f"Typographic {name}"

    space = glyphsLib.to_designspace(file, minimal=True)
    for instance, name in zip(space.instances, ["Thin", "Regular", "Black"]):
        source = copy.deepcopy(space.sources[0])
        apply_instance_data_to_ufo(source.font, instance, space)

        actual = source.font.info.openTypeNamePreferredFamilyName
        assert actual == "Typographic Neue"

        actual = source.font.info.openTypeNamePreferredSubfamilyName
        assert actual == f"Typographic {name}"



# Copyright 2015 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import logging

from fontTools.varLib.models import piecewiseLinearMap

from glyphsLib import classes
from glyphsLib.classes import WEIGHT_CODES, WIDTH_CODES, InstanceType
from glyphsLib.builder.constants import WIDTH_CLASS_TO_VALUE
from glyphsLib.types import parse_float_or_int

logger = logging.getLogger(__name__)


def class_to_value(axis, ufo_class):
    """
    >>> class_to_value('wdth', 7)
    125
    """
    if axis == "wght":
        # 600.0 => 600, 250 => 250
        return int(ufo_class)
    elif axis == "wdth":
        return WIDTH_CLASS_TO_VALUE[int(ufo_class)]

    raise NotImplementedError


def _nospace_lookup(dict, key):
    try:
        return dict[key]
    except KeyError:
        # Even though the Glyphs UI strings are supposed to be fixed,
        # some Noto files contain variants of them that have spaces.
        key = "".join(str(key).split())
        return dict[key]


def user_loc_string_to_value(axis_tag, user_loc):
    """Go from Glyphs UI strings to user space location.
    Returns None if the string is invalid.

    >>> user_loc_string_to_value('wght', 'ExtraLight')
    200
    >>> user_loc_string_to_value('wdth', 'SemiCondensed')
    87.5
    >>> user_loc_string_to_value('wdth', 'Clearly Not From Glyphs UI')
    """
    if axis_tag == "wght":
        if isinstance(user_loc, int):
            value = user_loc
        else:
            try:
                value = _nospace_lookup(WEIGHT_CODES, user_loc)
            except KeyError:
                return None
        return class_to_value("wght", value)
    elif axis_tag == "wdth":
        if isinstance(user_loc, int):
            value = user_loc
        else:
            try:
                value = _nospace_lookup(WIDTH_CODES, user_loc)
            except KeyError:
                return None
        return class_to_value("wdth", value)

    # Currently this function should only be called with a width or weight
    raise NotImplementedError


def user_loc_value_to_class(axis_tag, user_loc):
    """Return the OS/2 weight or width class that is closest to the provided
    user location. For weight the user location is between 0 and 1000 and for
    width it is a percentage.

    >>> user_loc_value_to_class('wght', 310)
    310
    >>> user_loc_value_to_class('wdth', 62)
    2
    """
    if axis_tag == "wght":
        return int(user_loc)
    elif axis_tag == "wdth":
        return min(
            sorted(WIDTH_CLASS_TO_VALUE.items()),
            key=lambda item: abs(item[1] - user_loc),
        )[0]

    raise NotImplementedError


def user_loc_value_to_instance_string(axis_tag, user_loc):
    """Return the Glyphs UI string (from the instance dropdown) that is
    closest to the provided user location.

    >>> user_loc_value_to_instance_string('wght', 430)
    'Normal'
    >>> user_loc_value_to_instance_string('wdth', 150)
    'Extra Expanded'
    """
    codes = {}
    if axis_tag == "wght":
        codes = WEIGHT_CODES
    elif axis_tag == "wdth":
        codes = WIDTH_CODES
    else:
        raise NotImplementedError
    class_ = user_loc_value_to_class(axis_tag, user_loc)
    return min(
        sorted((code, class_) for code, class_ in codes.items() if code is not None),
        key=lambda item: abs(item[1] - class_),
    )[0]


def update_mapping_from_instances(
    mapping, instances, axis_def, minimize_glyphs_diffs, cp_only=False
):
    # Collect the axis mappings from instances and update the mapping dict.
    for instance in instances:
        if instance.type == InstanceType.VARIABLE:
            continue
        if is_instance_active(instance) or minimize_glyphs_diffs:
            designLoc = axis_def.get_design_loc(instance)
            if cp_only:
                # Only use the Axis Location custom parameter for the user location
                userLoc = axis_def.get_user_loc_from_axis_location_cp(instance)
            else:
                # Use all heuristics to derive a user location
                userLoc = axis_def.get_user_loc(instance)
            if userLoc is None:
                # May happen if the custom parameter is disabled
                continue
            if userLoc in mapping and mapping[userLoc] != designLoc:
                logger.warning(
                    f"Axis {axis_def.tag}: Instance '{instance.name}' redefines "
                    f"the mapping for user location {userLoc} "
                    f"from {mapping[userLoc]} to {designLoc}"
                )
            mapping[userLoc] = designLoc


def is_identity(mapping):
    """Return whether the mapping is an identity mapping."""
    return all(userLoc == designLoc for userLoc, designLoc in mapping.items())


def to_designspace_axes(self):
    if not self.font.masters:
        return
    regular_master = get_regular_master(self.font)
    assert isinstance(regular_master, classes.GSFontMaster)

    custom_mapping = self.font.customParameters["Axis Mappings"]
    virtual_masters = [
        {v["Axis"]: v["Location"] for v in cp.value}
        for cp in self.font.customParameters
        if cp.name == "Virtual Master"
        if not cp.disabled
    ]

    for axis_def in get_axis_definitions(self.font):
        axis = self.designspace.newAxisDescriptor()
        axis.tag = axis_def.tag
        axis.name = axis_def.name
        axis.hidden = axis_def.hidden
        # TODO add support for localised axis.labelNames when Glyphs.app does

        # Make sure here that if there is an Axes parameter, we actually really
        # keep axes that are defined in it, even if they do nothing. This
        # prevents an edge case where you have a single master at the default
        # location that ends in with a Designspace with no axes, as they were
        # all filtered out by the `if` statement at the bottom.
        axis_wanted = False
        if any(
            a.get("Tag") == axis.tag for a in self.font.customParameters["Axes"] or []
        ):
            axis_wanted = True

        # See https://github.com/googlefonts/glyphsLib/issues/568
        if custom_mapping:
            if axis.tag in custom_mapping:
                mapping = {float(k): v for k, v in custom_mapping[axis.tag].items()}
                regularDesignLoc = axis_def.get_design_loc(regular_master)
                reverse_mapping = {dl: ul for ul, dl in sorted(mapping.items())}
                regularUserLoc = piecewiseLinearMap(regularDesignLoc, reverse_mapping)
            else:
                logger.debug(
                    f"Skipping {axis.tag} since it hasn't been defined "
                    "in the Axis Mapping."
                )
                continue
        # See https://github.com/googlefonts/glyphsLib/issues/280
        elif font_uses_axis_locations(self.font):
            # If all masters have an "Axis Location" custom parameter, only the values
            # from this parameter will be used to build the mapping of the masters and
            # instances.
            mapping = {}
            for master in self.font.masters:
                designLoc = axis_def.get_design_loc(master)
                userLoc = axis_def.get_user_loc(master)
                if userLoc in mapping and mapping[userLoc] != designLoc:
                    logger.warning(
                        f"Axis {axis_def.tag}: Master '{master.name}' redefines "
                        f"the mapping for user location {userLoc} "
                        f"from {mapping[userLoc]} to {designLoc}"
                    )
                mapping[userLoc] = designLoc

            update_mapping_from_instances(
                mapping,
                self.font.instances,
                axis_def,
                minimize_glyphs_diffs=self.minimize_glyphs_diffs,
                # Glyphs doesn't deduce instance mappings if font uses axis locations.
                # Use only the custom parameter if present.
                cp_only=True,
            )

            regularDesignLoc = axis_def.get_design_loc(regular_master)
            regularUserLoc = axis_def.get_user_loc(regular_master)
        else:
            # Build the mapping from the instances because they have both
            # a user location and a design location.
            instance_mapping = {}
            update_mapping_from_instances(
                instance_mapping,
                self.font.instances,
                axis_def,
                minimize_glyphs_diffs=self.minimize_glyphs_diffs,
            )

            master_mapping = {}
            for master in self.font.masters:
                # Glyphs masters don't have a user location
                userLoc = designLoc = axis_def.get_design_loc(master)
                master_mapping[userLoc] = designLoc

            # Prefer the instance-based mapping (but only if interesting)
            mapping = (
                instance_mapping
                if (instance_mapping and not is_identity(instance_mapping))
                else master_mapping
            )

            regularDesignLoc = axis_def.get_design_loc(regular_master)
            # Glyphs masters don't have a user location, so we compute it by
            # looking at the axis mapping in reverse.
            reverse_mapping = {dl: ul for ul, dl in sorted(mapping.items())}
            regularUserLoc = piecewiseLinearMap(regularDesignLoc, reverse_mapping)
            # TODO make sure that the default is in mapping?

        is_identity_map = is_identity(mapping)

        # Virtual Masters can't have an Axis Location parameter; their coordinates
        # can either be mapped via Axis Mappings, or implicitly by neighbouring
        # non-virtual masters' Axis Location params at least for existing axes; for
        # newly defined axes the virtual master coordinates are assumed to be un-mapped
        # (user==design).
        # Only if the {user:design} mapping so far is an identity map (because it
        # has not been 'bent' by one of the above mechanisms), the virtual masters
        # contribute to extend the current axis' min/max range.
        # https://github.com/googlefonts/glyphsLib/issues/859
        if is_identity_map:
            for vm in virtual_masters:
                for axis_name, axis_coord in vm.items():
                    if axis_name != axis.name:
                        continue
                    mapping[axis_coord] = axis_coord

        minimum = min(mapping)
        maximum = max(mapping)
        default = min(maximum, max(minimum, regularUserLoc))  # clamp

        if (
            minimum < maximum
            or minimum != axis_def.default_user_loc
            or not is_identity_map
            or axis_wanted
        ):
            if not is_identity_map:
                axis.map = sorted(mapping.items())
            axis.minimum = minimum
            axis.maximum = maximum
            axis.default = default
            self.designspace.addAxis(axis)

    # If there are no interesting axes, but only a single master at default location
    # along all 3 predefined axes, all with identity user:design mapping, we end up
    # with an empty list of axes, which is invalid. Thus as last resort we emit a
    # do-nothing Weight axis (the default axis when no "Axes" custom parameter is
    # defined) where default==min==max==400.
    # https://github.com/googlefonts/fontmake/issues/644
    if not self.designspace.axes:
        self.designspace.addAxisDescriptor(
            name=WEIGHT_AXIS_DEF.name,
            tag=WEIGHT_AXIS_DEF.tag,
            minimum=WEIGHT_AXIS_DEF.default_user_loc,
            default=WEIGHT_AXIS_DEF.default_user_loc,
            maximum=WEIGHT_AXIS_DEF.default_user_loc,
        )


def font_uses_axis_locations(font):
    # It's possible for fonts to have the 'Axes' parameter but to NOT specify
    # the master locations using 'Axis Location', in which case we have to
    # resort to using instances or other old tricks to get the mapping.
    # https://github.com/googlefonts/glyphsLib/issues/409
    # https://github.com/googlefonts/glyphsLib/issues/411
    return font.axes and all(
        master.customParameters["Axis Location"] for master in font.masters
    )


def to_glyphs_axes(self):
    axes_parameter = []
    for axis in self.designspace.axes:
        if axis.tag == "wght":
            name = axis.name or "Weight"
        elif axis.tag == "wdth":
            name = axis.name or "Width"
        else:
            name = axis.name
        axes_parameter.append(
            classes.GSAxis(name=name, tag=axis.tag, hidden=axis.hidden)
        )

    if axes_parameter and not _is_subset_of_default_axes(axes_parameter):
        self.font.axes = axes_parameter

    if any(_has_meaningful_map(a, self.designspace) for a in self.designspace.axes):
        mapping = {
            axis.tag: {str(k): v for k, v in axis.map} for axis in self.designspace.axes
        }
        self.font.customParameters["Axis Mappings"] = mapping


class AxisDefinition:
    """Centralize the code that deals with axis locations, user location versus
    design location, associated OS/2 table codes, etc.
    """

    def __init__(
        self,
        tag,
        name,
        design_loc_key,
        default_design_loc=0.0,
        user_loc_key=None,
        user_loc_param=None,
        default_user_loc=0.0,
        hidden=False,
    ):
        self.tag = tag
        self.name = name
        self.design_loc_key = design_loc_key
        self.default_design_loc = default_design_loc
        self.user_loc_key = user_loc_key
        self.user_loc_param = user_loc_param
        self.default_user_loc = default_user_loc
        self.hidden = hidden

    def get_design_loc(self, glyphs_master_or_instance):
        """Get the design location (aka interpolation value) of a Glyphs
        master or instance along this axis. For example for the weight
        axis it could be the thickness of a stem, for the width a percentage
        of extension with respect to the normal width.
        """
        return glyphs_master_or_instance._get_axis_value(self.design_loc_key)

    def set_design_loc(self, master_or_instance, value):
        """Set the design location of a Glyphs master or instance."""
        master_or_instance._set_axis_value(self.design_loc_key, value)

    def get_user_loc(self, master_or_instance):
        """Get the user location of a Glyphs master or instance.
        Masters and instances in Glyphs can have a user location in the
        "Axis Location" custom parameter.

        The user location is what the user sees on the slider in his
        variable-font-enabled UI. For weight it is a value between 0 and 1000,
        400 being Regular and 700 Bold.

        For width it's a percentage of extension with respect to the normal
        width, 100 being normal, 200 Ultra-expanded = twice as wide.
        It may or may not match the design location.
        """
        user_loc = self.default_user_loc

        if self.tag != "wght":
            # The user location is by default the same as the design location.
            user_loc = self.get_design_loc(master_or_instance)

        # Try to guess the user location by looking at the OS/2 weightClass
        # and widthClass. If a weightClass is found, it translates directly
        # to a user location in 0..1000. If a widthClass is found, it
        # translate to a percentage of extension according to the spec, see
        # the mapping named `WIDTH_CLASS_TO_VALUE` at the top.
        if self.user_loc_key is not None and hasattr(
            master_or_instance, self.user_loc_key
        ):
            # Instances have special ways to specify a user location.
            # Only weight and with have a custom user location via a key.
            # The `user_loc_key` gives a "location code" = Glyphs UI string
            user_loc_str = getattr(master_or_instance, self.user_loc_key)
            new_user_loc = user_loc_string_to_value(self.tag, user_loc_str)
            if new_user_loc is not None:
                user_loc = new_user_loc

        # The custom param takes over the key if it exists
        # e.g. for weight:
        #       key = "weight" -> "Bold" -> 700
        # but param = "weightClass" -> 600       => 600 wins
        if self.user_loc_param is not None:
            class_ = master_or_instance.customParameters[self.user_loc_param]
            if class_ is not None:
                user_loc = class_to_value(self.tag, class_)

        axis_location = self.get_user_loc_from_axis_location_cp(master_or_instance)
        if axis_location is not None:
            user_loc = axis_location

        return user_loc

    def get_user_loc_from_axis_location_cp(self, master_or_instance):
        # Masters and instances have a customParameter that specifies a user location
        # along custom axes. If this is present it takes precedence over everything
        # else.
        loc_param = master_or_instance.customParameters["Axis Location"]
        try:
            for location in loc_param:
                if location.get("Axis") == self.name:
                    return parse_float_or_int(location["Location"])
        except (TypeError, KeyError):
            pass

    def set_user_loc(self, master_or_instance, value):
        """Set the user location of a Glyphs master or instance."""
        if isinstance(master_or_instance, classes.GSInstance):
            # The following code is only valid for instances.
            # Masters also the keys `weight` and `width` but they should not be
            # used, they are deprecated and should only be used to store
            # (parts of) the master's name, but not its location.

            # Try to set the key if possible, i.e. if there is a key, and
            # if there exists a code that can represent the given value, e.g.
            # for "weight": 600 can be represented by SemiBold so we use that,
            # but for 550 there is no code so we will have to set the custom
            # parameter as well.
            if self.user_loc_key is not None and hasattr(
                master_or_instance, self.user_loc_key
            ):
                code = user_loc_value_to_instance_string(self.tag, value)
                value_for_code = user_loc_string_to_value(self.tag, code)
                setattr(master_or_instance, self.user_loc_key, code)
                if self.user_loc_param is not None and value != value_for_code:
                    try:
                        class_ = user_loc_value_to_class(self.tag, value)
                        master_or_instance.customParameters[self.user_loc_param] = (
                            class_
                        )
                    except NotImplementedError:
                        # user_loc_value_to_class only works for weight & width
                        pass
            return

        # Directly set the custom parameter (old way) and also the Axis Location
        # (new way).
        if self.user_loc_param is not None:
            try:
                class_ = user_loc_value_to_class(self.tag, value)
                master_or_instance.customParameters[self.user_loc_param] = class_
            except NotImplementedError:
                pass

        loc_param = master_or_instance.customParameters["Axis Location"]
        if loc_param is None:
            loc_param = []
            master_or_instance.customParameters["Axis Location"] = loc_param
        location = None
        for loc in loc_param:
            if loc.get("Axis") == self.name:
                location = loc
        if location is None:
            loc_param.append({"Axis": self.name, "Location": value})
        else:
            location["Location"] = value

    def set_user_loc_code(self, instance, code):
        assert isinstance(instance, classes.GSInstance)
        # The previous method `set_user_loc` will not roundtrip every
        # time, for example for value = 600, both "DemiBold" and "SemiBold"
        # would work, so we provide this other method to set a specific code.
        if self.user_loc_key is not None:
            setattr(instance, self.user_loc_key, code)

    def set_ufo_user_loc(self, ufo, value):
        if self.tag not in ("wght", "wdth"):
            raise NotImplementedError
        class_ = user_loc_value_to_class(self.tag, value)
        ufo_key = (
            "openTypeOS2WeightClass" if self.tag == "wght" else "openTypeOS2WidthClass"
        )
        setattr(ufo.info, ufo_key, class_)


class AxisDefinitionFactory:
    """Creates a set of axis definitions, making sure to recognize default axes
    (weight and width) and also keeping track of indices of custom axes.

    From looking at a Glyphs file with only one custom axis, it looks like
    when there is an "Axes" customParameter, the axis design locations are
    stored in `weightValue` for the first axis (regardless of whether it is
    a weight axis, `widthValue` for the second axis, etc.
    """

    def __init__(self):
        self.axis_index = -1

    def get(self, tag=None, name="Custom", hidden=False):
        self.axis_index += 1
        design_loc_key = self._design_loc_key()
        if tag is None:
            if self.axis_index == 0:
                tag = "XXXX"
            else:
                tag = "XXX%d" % self.axis_index

        if tag == "wght":
            default_design_loc, default_user_loc = 100.0, 400.0
            user_loc_key, user_loc_param = "weight", "weightClass"
        elif tag == "wdth":
            default_design_loc = default_user_loc = 100.0
            user_loc_key, user_loc_param = "width", "widthClass"
        else:
            default_design_loc = default_user_loc = 0.0
            user_loc_key = user_loc_param = None

        return AxisDefinition(
            tag,
            name,
            design_loc_key,
            default_design_loc=default_design_loc,
            user_loc_key=user_loc_key,
            user_loc_param=user_loc_param,
            default_user_loc=default_user_loc,
            hidden=hidden,
        )

    def _design_loc_key(self):
        return self.axis_index


defaults_factory = AxisDefinitionFactory()

WEIGHT_AXIS_DEF = defaults_factory.get("wght", "Weight")
WIDTH_AXIS_DEF = defaults_factory.get("wdth", "Width")
CUSTOM_AXIS_DEF = defaults_factory.get("XXXX", "Custom")
DEFAULT_AXES_DEFS = (WEIGHT_AXIS_DEF, WIDTH_AXIS_DEF, CUSTOM_AXIS_DEF)


# Adapted from PR https://github.com/googlefonts/glyphsLib/pull/306
def get_axis_definitions(font):
    axesParameter = font.customParameters["Axes"]
    if axesParameter is None:
        return DEFAULT_AXES_DEFS

    factory = AxisDefinitionFactory()
    return [
        factory.get(axis.get("Tag"), axis["Name"], bool(axis.get("Hidden")))
        for axis in axesParameter
    ]


def _is_subset_of_default_axes(axes_parameter):
    if len(axes_parameter) > 3:
        return False
    for axis, axis_def in zip(axes_parameter, DEFAULT_AXES_DEFS):
        if axis.name != axis_def.name:
            return False
        if axis.axisTag != axis_def.tag:
            return False
    return True


def _has_meaningful_map(axis, designspace):
    if not axis.map:
        return False
    for k, v in axis.map:
        if k != v:
            return True
    # We have an identity map. We could elide it, but...
    # sometimes we use an identity map to force a particular
    # range even though the sources don't fill that range.
    min_axis = None
    max_axis = None
    for source in designspace.sources:
        loc = source.location.get(axis.name)
        if loc is None:
            continue
        if min_axis is None:
            min_axis = loc
        else:
            min_axis = min(loc, min_axis)
        if max_axis is None:
            max_axis = loc
        else:
            max_axis = max(loc, max_axis)
    if min_axis != axis.map[0][0] or max_axis != axis.map[-1][0]:
        return True
    return False


def get_regular_master(font):
    """Find the "regular" master among the GSFontMasters.

    Tries to find the master with the passed 'regularName'.
    If there is no such master or if regularName is None,
    tries to find a base style shared between all masters
    (defaulting to "Regular"), and then tries to find a master
    with that style name. If there is no master with that name,
    returns the first master in the list.
    """
    if not font.masters:
        return None
    # The current glyphs source specification supports the custom
    # parameter name "Variable Font Origin".  This may have been
    # named "Variation Font Origin" in the past.
    # We support the current name with a fallback to the previous name
    # if not found in the GSFont.customParameters dict
    if "Variable Font Origin" in font.customParameters:
        regular_id = font.customParameters["Variable Font Origin"]
        if regular_id:
            for master in font.masters:
                if master.id == regular_id:
                    return master
    elif "Variation Font Origin" in font.customParameters:
        regular_name = font.customParameters["Variation Font Origin"]
        if regular_name:
            for master in font.masters:
                if master.name == regular_name:
                    return master
    base_style = find_base_style(font.masters)
    if not base_style:
        base_style = "Regular"
    for master in font.masters:
        if master.name == base_style:
            return master
    # Second try: maybe the base style has regular in it as well
    for master in font.masters:
        name_without_regular = " ".join(
            n for n in master.name.split(" ") if n != "Regular"
        )
        if name_without_regular == base_style:
            return master
    return font.masters[0]


def find_base_style(masters):
    """Find a base style shared between all masters.
    Return empty string if none is found.
    """
    if not masters:
        return ""
    base_style = (masters[0].name or "").split()
    for master in masters:
        style = master.name.split()
        base_style = [s for s in style if s in base_style]
    base_style = " ".join(base_style)
    return base_style


def is_instance_active(instance):
    # Glyphs.app recognizes both "exports=0" and "active=0" as a flag
    # to mark instances as inactive. Inactive instances should get ignored.
    # https://github.com/googlefonts/glyphsLib/issues/129
    return instance.exports and getattr(instance, "active", True)



#
# Copyright 2017 Google Inc. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import os.path
import shutil
import sys
import tempfile
import unittest
import xml.etree.ElementTree as etree
from xmldiff import main, formatting

import defcon
from glyphsLib.builder.constants import GLYPHS_PREFIX
from glyphsLib.builder.instances import set_weight_class, set_width_class
from glyphsLib.classes import GSFont, GSFontMaster, GSInstance, GSFontInfoValue
from glyphsLib import to_designspace, to_glyphs
import ufoLib2


# Current limitation of glyphsLib for designspace to designspace round-trip:
# the designspace's axes, sources and instances must be as such:
#  - the axes' min, max must match extreme masters
#  - the axes' default must match the "regular master"
#  - the axes' mapping must have as many pairs as instances, each pair
#    matching the (userLoc, designLoc) of an instance. If there are no
#    instances, same requirement but with 1 pair/master
#
# Basically this is to say that the designspace must have been generated by
# glyphsLib in the first place.
#
# More general designspaces (like: axes that go farther than extreme masters,
# mapping with arbitrary numbers of pairs...) might be supported later, if/when
# Glyphs gets a UI to setup this information.
#
# REVIEW: check that the above makes sense


def makeFamily():
    m1 = makeMaster("Regular", weight=90.0)
    m2 = makeMaster("Black", weight=190.0)
    instances = [
        makeInstance("Regular", weight=("Regular", 400, 90)),
        makeInstance("Semibold", weight=("SemiBold", 600, 128)),
        makeInstance("Bold", weight=("Bold", 700, 151), is_bold=True),
        makeInstance("Black", weight=("Black", 900, 190)),
    ]
    return [m1, m2], instances


def makeMaster(styleName, weight=None, width=None):
    m = GSFontMaster()
    m.name = styleName
    if weight is not None:
        m.weightValue = weight
    if width is not None:
        m.widthValue = width
    return m


def makeInstance(
    name, weight=None, width=None, is_bold=None, is_italic=None, linked_style=None
):
    inst = GSInstance()
    inst.name = name
    if weight is not None:
        # Glyphs 2.5 stores the instance weight in two to three places:
        # 1. as a textual weight (such as “Bold”; no value defaults to
        #    "Regular");
        # 2. (optional) as numeric customParameters.weightClass (such as 700),
        #    which corresponds to OS/2.usWeightClass where 100 means Thin,
        #    400 means Regular, 700 means Bold, and 900 means Black;
        # 3. as numeric weightValue (such as 66.0), which typically is
        #    the stem width but can be anything that works for interpolation
        #    (no value defaults to 100).
        weightName, weightClass, interpolationWeight = weight
        if weightName is not None:
            inst.weight = weightName
        if weightClass is not None:
            inst.customParameters["weightClass"] = weightClass
        if interpolationWeight is not None:
            inst.weightValue = interpolationWeight
    if width is not None:
        # Glyphs 2.5 stores the instance width in two to three places:
        # 1. as a textual width (such as “Condensed”; no value defaults
        #    to "Medium (normal)");
        # 2. (optional) as numeric customParameters.widthClass (such as 5),
        #    which corresponds to OS/2.usWidthClass where 1 means Ultra-
        #    condensed, 5 means Medium (normal), and 9 means Ultra-expanded;
        # 3. as numeric widthValue (such as 79), which typically is
        #    a percentage of whatever the font designer considers “normal”
        #    but can be anything that works for interpolation (no value
        #    defaults to 100).
        widthName, widthClass, interpolationWidth = width
        if widthName is not None:
            inst.width = widthName
        if widthClass is not None:
            inst.customParameters["widthClass"] = widthClass
        if interpolationWidth is not None:
            inst.widthValue = interpolationWidth
    # TODO: Support custom axes; need to triple-check how these are encoded in
    # Glyphs files. Glyphs 3 will likely overhaul the representation of axes.
    if is_bold is not None:
        inst.isBold = is_bold
    if is_italic is not None:
        inst.isItalic = is_italic
    if linked_style is not None:
        inst.linkStyle = linked_style
    return inst


def makeInstanceDescriptor(*args, **kwargs):
    """Same as makeInstance but return the corresponding InstanceDescriptor."""
    ginst = makeInstance(*args, **kwargs)
    font = makeFont([makeMaster("Regular")], [ginst], "Family")
    doc = to_designspace(font)
    return doc, doc.instances[0]


def makeFont(masters, instances, familyName):
    font = GSFont()
    font.familyName = familyName
    font.masters = masters
    font.instances = instances
    return font


class DesignspaceTest(unittest.TestCase):
    def setUp(self):
        self.tmpdir = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.tmpdir)

    def write_to_tmp_path(self, doc, name):
        path = os.path.join(self.tmpdir, name)
        doc.write(path)
        return path

    def expect_designspace(self, doc, expected_name):
        dirname = os.path.dirname(__file__)
        expected_path = os.path.join(dirname, "..", "data", expected_name)
        return self._expect_designspace(doc, expected_path)

    def _expect_designspace(self, doc, expected_path):
        actual_path = self.write_to_tmp_path(doc, "generated.designspace")
        actual_diff = main.diff_files(
            actual_path, expected_path, formatter=formatting.DiffFormatter()
        )
        if len(actual_diff) != 0:
            expected_name = os.path.basename(expected_path)
            sys.stderr.write("%s discrepancies (per xmldiff):\n" % (expected_name))
            for line in actual_diff.split("\n"):
                sys.stderr.write("  %s" % (line))
            self.fail("*.designspace file is different from expected")

    def expect_designspace_roundtrip(self, doc):
        actual_path = self.write_to_tmp_path(doc, "original.designspace")
        font = to_glyphs(doc, minimize_ufo_diffs=True)
        rtdoc = to_designspace(font)
        return self._expect_designspace(rtdoc, actual_path)

    def test_basic(self):
        masters, instances = makeFamily()
        font = makeFont(masters, instances, "DesignspaceTest Basic")
        doc = to_designspace(font, instance_dir="out")
        self.expect_designspace(doc, "DesignspaceTestBasic.designspace")
        self.expect_designspace_roundtrip(doc)

    def test_inactive_from_exports(self):
        # Glyphs.app recognizes exports=0 as a flag for inactive instances.
        # https://github.com/googlefonts/glyphsLib/issues/129
        masters, instances = makeFamily()
        for inst in instances:
            if inst.name != "Semibold":
                inst.exports = False
        font = makeFont(masters, instances, "DesignspaceTest Inactive")
        doc = to_designspace(font, instance_dir="out")
        self.expect_designspace(doc, "DesignspaceTestInactive.designspace")
        self.expect_designspace_roundtrip(doc)

        # Although inactive instances are not exported by default,
        # all instances are exported when intending to roundtrip Glyphs->Glyphs
        doc = to_designspace(font, minimize_glyphs_diffs=True)
        self.assertEqual(4, len(doc.instances))

    def test_familyName(self):
        masters, _ = makeFamily()
        customFamily = makeInstance("Regular", weight=("Bold", 600, 151))
        customFamily.customParameters["familyName"] = "Custom Family"
        instances = [makeInstance("Regular", weight=("Regular", 400, 90)), customFamily]
        font = makeFont(masters, instances, "DesignspaceTest FamilyName")
        doc = to_designspace(font, instance_dir="out")
        self.expect_designspace(doc, "DesignspaceTestFamilyName.designspace")
        self.expect_designspace_roundtrip(doc)

    def test_fileName(self):
        masters, _ = makeFamily()
        customFileName = makeInstance("Regular", weight=("Bold", 600, 151))
        customFileName.customParameters["fileName"] = "Custom FileName"
        instances = [
            makeInstance("Regular", weight=("Regular", 400, 90)),
            customFileName,
        ]
        font = makeFont(masters, instances, "DesignspaceTest FamilyName")
        doc = to_designspace(font, instance_dir="out")
        self.expect_designspace(doc, "DesignspaceTestFileName.designspace")
        self.expect_designspace_roundtrip(doc)

    def test_noRegularMaster(self):
        # Currently, fontTools.varLib fails to build variable fonts
        # if the default axis value does not happen to be at the
        # location of one of the interpolation masters.
        # glyhpsLib tries to work around this downstream limitation.
        masters = [makeMaster("Thin", weight=26), makeMaster("Black", weight=190)]
        instances = [
            makeInstance("Black", weight=("Black", 900, 190)),
            makeInstance("Regular", weight=("Regular", 400, 90)),
            makeInstance("Bold", weight=("Thin", 100, 26)),
        ]
        font = makeFont(masters, instances, "NoRegularMaster")
        designspace = to_designspace(font, instance_dir="out")
        path = self.write_to_tmp_path(designspace, "noregular.designspace")
        doc = etree.parse(path)
        weightAxis = doc.find('axes/axis[@tag="wght"]')
        self.assertEqual(weightAxis.attrib["minimum"], "100")
        self.assertEqual(weightAxis.attrib["default"], "100")  # not 400
        self.assertEqual(weightAxis.attrib["maximum"], "900")

        self.expect_designspace_roundtrip(designspace)

    def test_postscriptFontNameCustomParameter(self):
        master = makeMaster("Master")
        thin, black = makeInstance("Thin"), makeInstance("Black")
        black.customParameters["postscriptFontName"] = "PSNameTest-Superfat"
        font = makeFont([master], [thin, black], "PSNameTest")
        designspace = to_designspace(font, instance_dir="out")
        path = self.write_to_tmp_path(designspace, "psname.designspace")
        d = etree.parse(path)

        def psname(doc, style):
            inst = doc.find('instances/instance[@stylename="%s"]' % style)
            return inst.attrib.get("postscriptfontname")

        self.assertIsNone(psname(d, "Thin"))
        self.assertEqual(psname(d, "Black"), "PSNameTest-Superfat")

        self.expect_designspace_roundtrip(designspace)

    def test_postscriptFontNameProperty(self):
        master = makeMaster("Master")
        thin, black = makeInstance("Thin"), makeInstance("Black")
        black.properties.append(
            GSFontInfoValue("postscriptFontName", "PSNameTest-Superfat")
        )
        font = makeFont([master], [thin, black], "PSNameTest")
        designspace = to_designspace(font, instance_dir="out")
        path = self.write_to_tmp_path(designspace, "psname.designspace")
        d = etree.parse(path)

        def psname(doc, style):
            inst = doc.find('instances/instance[@stylename="%s"]' % style)
            return inst.attrib.get("postscriptfontname")

        self.assertIsNone(psname(d, "Thin"))
        self.assertEqual(psname(d, "Black"), "PSNameTest-Superfat")

        self.expect_designspace_roundtrip(designspace)

    def test_variablePostscriptFontNameProperty(self):
        master = makeMaster("Master")
        thin, black = makeInstance("Thin"), makeInstance("Black")
        black.properties.append(
            GSFontInfoValue("variablePostscriptFontName", "PSNameTest-Superfat")
        )
        font = makeFont([master], [thin, black], "PSNameTest")
        designspace = to_designspace(font, instance_dir="out")
        path = self.write_to_tmp_path(designspace, "psname.designspace")
        d = etree.parse(path)

        def psname(doc, style):
            inst = doc.find('instances/instance[@stylename="%s"]' % style)
            return inst.attrib.get("postscriptfontname")

        self.assertIsNone(psname(d, "Thin"))
        self.assertEqual(psname(d, "Black"), "PSNameTest-Superfat")

        self.expect_designspace_roundtrip(designspace)

    def test_instanceOrder(self):
        # The generated *.designspace file should place instances
        # in the same order as they appear in the original source.
        # https://github.com/googlefonts/glyphsLib/issues/113
        masters, _ = makeFamily()
        instances = [
            makeInstance("Black", weight=("Black", 900, 190)),
            makeInstance("Regular", weight=("Regular", 400, 90)),
            makeInstance("Bold", weight=("Bold", 700, 151), is_bold=True),
        ]
        font = makeFont(masters, instances, "DesignspaceTest InstanceOrder")
        doc = to_designspace(font, instance_dir="out")
        self.expect_designspace(doc, "DesignspaceTestInstanceOrder.designspace")
        self.expect_designspace_roundtrip(doc)

    def test_twoAxes(self):
        # In NotoSansArabic-MM.glyphs, the regular width only contains
        # parameters for the weight axis. For the width axis, glyphsLib
        # should use 100 as default value (just like Glyphs.app does).
        familyName = "DesignspaceTest TwoAxes"
        masters = [
            makeMaster("Regular", weight=90),
            makeMaster("Black", weight=190),
            makeMaster("Thin", weight=26),
            makeMaster("ExtraCond", weight=90, width=70),
            makeMaster("ExtraCond Black", weight=190, width=70),
            makeMaster("ExtraCond Thin", weight=26, width=70),
        ]
        instances = [
            makeInstance("Thin", weight=("Thin", 100, 26)),
            makeInstance("Regular", weight=("Regular", 400, 90)),
            makeInstance("Semibold", weight=("SemiBold", 600, 128)),
            makeInstance("Black", weight=("Black", 900, 190)),
            makeInstance(
                "ExtraCondensed Thin",
                weight=("Thin", 100, 26),
                width=("Extra Condensed", 2, 70),
            ),
            makeInstance(
                "ExtraCondensed",
                weight=("Regular", 400, 90),
                width=("Extra Condensed", 2, 70),
            ),
            makeInstance(
                "ExtraCondensed Black",
                weight=("Black", 900, 190),
                width=("Extra Condensed", 2, 70),
            ),
        ]
        font = makeFont(masters, instances, familyName)
        doc = to_designspace(font, instance_dir="out")
        self.expect_designspace(doc, "DesignspaceTestTwoAxes.designspace")
        self.expect_designspace_roundtrip(doc)

    def test_variationFontOrigin(self):
        # Glyphs 2.4.1 introduced a custom parameter “Variation Font Origin”
        # to specify which master should be considered the origin.
        # https://glyphsapp.com/blog/glyphs-2-4-1-released
        masters = [
            makeMaster("Thin", weight=26),
            makeMaster("Regular", weight=100),
            makeMaster("Medium", weight=111),
            makeMaster("Black", weight=190),
        ]
        instances = [
            makeInstance("Black", weight=("Black", 900, 190)),
            makeInstance("Medium", weight=("Medium", 444, 111)),
            makeInstance("Regular", weight=("Regular", 400, 100)),
            makeInstance("Thin", weight=("Thin", 100, 26)),
        ]
        font = makeFont(masters, instances, "Family")
        font.customParameters["Variation Font Origin"] = "Medium"
        designspace = to_designspace(font, instance_dir="out")
        path = self.write_to_tmp_path(designspace, "varfontorig.designspace")
        doc = etree.parse(path)
        medium = doc.find('sources/source[@stylename="Medium"]')
        self.assertEqual(medium.find("lib").attrib["copy"], "1")
        weightAxis = doc.find('axes/axis[@tag="wght"]')
        self.assertEqual(weightAxis.attrib["default"], "444")

        self.expect_designspace_roundtrip(designspace)

    def test_designspace_name(self):
        doc = to_designspace(
            makeFont(
                [makeMaster("Regular", weight=100), makeMaster("Bold", weight=190)],
                [],
                "Family Name",
            )
        )
        # no shared base style name, only write the family name
        self.assertEqual(doc.filename, "FamilyName.designspace")

        doc = to_designspace(
            makeFont(
                [
                    makeMaster("Italic", weight=100),
                    makeMaster("Bold Italic", weight=190),
                ],
                [],
                "Family Name",
            )
        )
        # 'Italic' is the base style; append to designspace name
        self.assertEqual(doc.filename, "FamilyName-Italic.designspace")

    def test_instance_filtering_by_family_name(self):
        # See https://github.com/googlefonts/fontmake/issues/257
        path = os.path.join(
            os.path.dirname(__file__), "..", "data", "MontserratStrippedDown.glyphs"
        )
        font = GSFont(path)

        # By default (no special parameter), all instances are exported
        designspace_all = to_designspace(font)
        assert len(designspace_all.instances) == 18

        # If we specify that we want the same familyName as the masters,
        # we only get instances that have that same family name, and the
        # masters are copied as-is. (basically a subset of the previous doc)
        designspace_no_alternates = to_designspace(font, family_name="Montserrat")
        assert len(designspace_no_alternates.instances) == 9

        # If we specify the alternate family name, we only get the instances
        # that have that family name, and the masters are renamed to have the
        # given family name.
        designspace_alternates = to_designspace(
            font, family_name="Montserrat Alternates"
        )
        assert designspace_alternates.sources[0].familyName == "Montserrat Alternates"
        assert (
            designspace_alternates.sources[0].font.info.familyName
            == "Montserrat Alternates"
        )
        assert len(designspace_alternates.instances) == 9


WEIGHT_CLASS_KEY = GLYPHS_PREFIX + "weightClass"
WIDTH_CLASS_KEY = GLYPHS_PREFIX + "widthClass"


class SetWeightWidthClassesTestBase(object):
    ufo_module = None  # subclasses must override this

    def test_no_weight_class(self):
        ufo = self.ufo_module.Font()
        # name here says "Bold", however no explicit weightClass
        # is assigned
        doc, instance = makeInstanceDescriptor("Bold")
        set_weight_class(ufo, doc, instance)
        # the default OS/2 weight class is set
        self.assertEqual(ufo.info.openTypeOS2WeightClass, 400)

    def test_weight_class(self):
        ufo = self.ufo_module.Font()
        doc, data = makeInstanceDescriptor("Bold", weight=("Bold", None, 150))

        set_weight_class(ufo, doc, data)
        self.assertEqual(ufo.info.openTypeOS2WeightClass, 700)

    def test_explicit_default_weight(self):
        ufo = self.ufo_module.Font()
        doc, data = makeInstanceDescriptor("Regular", weight=("Regular", None, 100))

        set_weight_class(ufo, doc, data)
        # the default OS/2 weight class is set
        self.assertEqual(ufo.info.openTypeOS2WeightClass, 400)

    def test_no_width_class(self):
        ufo = self.ufo_module.Font()
        # no explicit widthClass set, instance name doesn't matter
        doc, data = makeInstanceDescriptor("Normal")
        set_width_class(ufo, doc, data)
        # the default OS/2 width class is set
        self.assertEqual(ufo.info.openTypeOS2WidthClass, 5)

    def test_width_class(self):
        ufo = self.ufo_module.Font()
        doc, data = makeInstanceDescriptor("Condensed", width=("Condensed", 3, 80))

        set_width_class(ufo, doc, data)
        self.assertEqual(ufo.info.openTypeOS2WidthClass, 3)

    def test_explicit_default_width(self):
        ufo = self.ufo_module.Font()
        doc, data = makeInstanceDescriptor("Regular", width=("Medium (normal)", 5, 100))

        set_width_class(ufo, doc, data)
        # the default OS/2 width class is set
        self.assertEqual(ufo.info.openTypeOS2WidthClass, 5)

    def test_weight_and_width_class(self):
        ufo = self.ufo_module.Font()
        doc, data = makeInstanceDescriptor(
            "SemiCondensed ExtraBold",
            weight=("ExtraBold", None, 160),
            width=("SemiCondensed", 4, 90),
        )

        set_weight_class(ufo, doc, data)
        set_width_class(ufo, doc, data)

        self.assertEqual(ufo.info.openTypeOS2WeightClass, 800)
        self.assertEqual(ufo.info.openTypeOS2WidthClass, 4)

    def test_unknown_ui_string_but_defined_weight_class(self):
        ufo = self.ufo_module.Font()
        # "DemiLight" is not among the predefined weight classes listed in
        # Glyphs.app/Contents/Frameworks/GlyphsCore.framework/Versions/A/
        # Resources/weights.plist
        # NOTE It is not possible from the user interface to set a custom
        # string as instance 'weightClass' since the choice is constrained
        # by a drop-down menu.
        doc, data = makeInstanceDescriptor(
            "DemiLight Italic", weight=("DemiLight", 350, 70)
        )

        set_weight_class(ufo, doc, data)

        # Here we have set the weightClass to 350 so even though the string
        # is wrong, our value of 350 should be used.
        self.assertTrue(ufo.info.openTypeOS2WeightClass == 350)

    def test_unknown_weight_class(self):
        ufo = self.ufo_module.Font()
        # "DemiLight" is not among the predefined weight classes listed in
        # Glyphs.app/Contents/Frameworks/GlyphsCore.framework/Versions/A/
        # Resources/weights.plist
        # NOTE It is not possible from the user interface to set a custom
        # string as instance 'weightClass' since the choice is constrained
        # by a drop-down menu.
        doc, data = makeInstanceDescriptor(
            "DemiLight Italic", weight=("DemiLight", None, 70)
        )

        set_weight_class(ufo, doc, data)

        # the default OS/2 weight class is set
        self.assertEqual(ufo.info.openTypeOS2WeightClass, 400)


class SetWeightWidthClassesTestUfoLib2(
    SetWeightWidthClassesTestBase, unittest.TestCase
):
    ufo_module = ufoLib2


class SetWeightWidthClassesTestDefcon(SetWeightWidthClassesTestBase, unittest.TestCase):
    ufo_module = defcon


if __name__ == "__main__":
    sys.exit(unittest.main())



from glyphsLib.builder.names import build_stylemap_names


def test_regular():
    map_family, map_style = build_stylemap_names(
        family_name="NotoSans",
        style_name="Regular",
        is_bold=False,
        is_italic=False,
        linked_style=None,
    )
    assert map_family == "NotoSans"
    assert map_style == "regular"


def test_regular_isBold():
    map_family, map_style = build_stylemap_names(
        family_name="NotoSans",
        style_name="Regular",
        is_bold=True,
        is_italic=False,
        linked_style=None,
    )
    assert map_family == "NotoSans Regular"
    assert map_style == "bold"


def test_regular_isItalic():
    map_family, map_style = build_stylemap_names(
        family_name="NotoSans",
        style_name="Regular",
        is_bold=False,
        is_italic=True,
        linked_style=None,
    )
    assert map_family == "NotoSans Regular"
    assert map_style == "italic"


def test_non_regular():
    map_family, map_style = build_stylemap_names(
        family_name="NotoSans",
        style_name="ExtraBold",
        is_bold=False,
        is_italic=False,
        linked_style=None,
    )
    assert map_family == "NotoSans ExtraBold"
    assert map_style == "regular"


def test_bold_no_style_link():
    map_family, map_style = build_stylemap_names(
        family_name="NotoSans",
        style_name="Bold",
        is_bold=False,  # not style-linked, despite the name
        is_italic=False,
        linked_style=None,
    )
    assert map_family == "NotoSans Bold"
    assert map_style == "regular"


def test_italic_no_style_link():
    map_family, map_style = build_stylemap_names(
        family_name="NotoSans",
        style_name="Italic",
        is_bold=False,
        is_italic=False,  # not style-linked, despite the name
        linked_style=None,
    )
    assert map_family == "NotoSans Italic"
    assert map_style == "regular"


def test_bold_italic_no_style_link():
    map_family, map_style = build_stylemap_names(
        family_name="NotoSans",
        style_name="Bold Italic",
        is_bold=False,  # not style-linked, despite the name
        is_italic=False,  # not style-linked, despite the name
        linked_style=None,
    )
    assert map_family == "NotoSans Bold Italic"
    assert map_style == "regular"


def test_bold():
    map_family, map_style = build_stylemap_names(
        family_name="NotoSans",
        style_name="Bold",
        is_bold=True,
        is_italic=False,
        linked_style=None,
    )
    assert map_family == "NotoSans"
    assert map_style == "bold"


def test_italic():
    map_family, map_style = build_stylemap_names(
        family_name="NotoSans",
        style_name="Italic",
        is_bold=False,
        is_italic=True,
        linked_style=None,
    )
    assert map_family == "NotoSans"
    assert map_style == "italic"


def test_bold_italic():
    map_family, map_style = build_stylemap_names(
        family_name="NotoSans",
        style_name="Bold Italic",
        is_bold=True,
        is_italic=True,
        linked_style=None,
    )
    assert map_family == "NotoSans"
    assert map_style == "bold italic"


def test_incomplete_bold_italic():
    map_family, map_style = build_stylemap_names(
        family_name="NotoSans",
        style_name="Bold",  # will be stripped...
        is_bold=True,
        is_italic=True,
        linked_style=None,
    )
    assert map_family == "NotoSans"
    assert map_style == "bold italic"

    map_family, map_style = build_stylemap_names(
        family_name="NotoSans",
        style_name="Italic",  # will be stripped...
        is_bold=True,
        is_italic=True,
        linked_style=None,
    )
    assert map_family == "NotoSans"
    assert map_style == "bold italic"


def test_italicbold_isBoldItalic():
    map_family, map_style = build_stylemap_names(
        family_name="NotoSans",
        style_name="Italic Bold",  # reversed
        is_bold=True,
        is_italic=True,
        linked_style=None,
    )
    assert map_family == "NotoSans"
    assert map_style == "bold italic"


def test_linked_style_regular():
    map_family, map_style = build_stylemap_names(
        family_name="NotoSans",
        style_name="Condensed",
        is_bold=False,
        is_italic=False,
        linked_style="Cd",
    )
    assert map_family == "NotoSans Cd"
    assert map_style == "regular"


def test_linked_style_bold():
    map_family, map_style = build_stylemap_names(
        family_name="NotoSans",
        style_name="Condensed Bold",
        is_bold=True,
        is_italic=False,
        linked_style="Cd",
    )
    assert map_family == "NotoSans Cd"
    assert map_style == "bold"


def test_linked_style_italic():
    map_family, map_style = build_stylemap_names(
        family_name="NotoSans",
        style_name="Condensed Italic",
        is_bold=False,
        is_italic=True,
        linked_style="Cd",
    )
    assert map_family == "NotoSans Cd"
    assert map_style == "italic"


def test_linked_style_bold_italic():
    map_family, map_style = build_stylemap_names(
        family_name="NotoSans",
        style_name="Condensed Bold Italic",
        is_bold=True,
        is_italic=True,
        linked_style="Cd",
    )
    assert map_family == "NotoSans Cd"
    assert map_style == "bold italic"



import copy
import logging
from collections import defaultdict


logger = logging.getLogger(__name__)


def align_alternate_layers(font, glyph_data=None):
    """Ensure composites have the same alternate layers as their components.

    If a glyph uses a component which has alternate (aka 'bracket') layers, that
    glyph also must have the same alternate layers or else it will not correctly swap
    when these are converted GSUB FeatureVariations.
    We copy the layer locations from the component into the glyph which uses it.
    """
    # First let's put all the layers in a sensible order so we can
    # query them efficiently
    master_layers = defaultdict(dict)
    alternate_layers = defaultdict(lambda: defaultdict(list))
    master_ids = set(master.id for master in font.masters)

    for glyph in font.glyphs:
        for layer in glyph.layers:
            if layer.layerId in master_ids:
                master_layers[layer.layerId][glyph.name] = layer
            elif layer.associatedMasterId in master_ids and layer._is_bracket_layer():
                alternate_layers[layer.associatedMasterId][glyph.name].append(layer)

    # Now let's find those which have a problem: they use components,
    # the components have some alternate layers, but the layer doesn't
    # have the same.
    # Because of the possibility of deeply nested components, we need
    # to keep doing this, bubbling up fixes until there's nothing left
    # to do.
    while True:
        problematic_glyphs = defaultdict(set)
        for master, layers in master_layers.items():
            for glyph_name, layer in layers.items():
                my_bracket_layers = {
                    tuple(layer._bracket_axis_rules())
                    for layer in alternate_layers[master][glyph_name]
                }
                for comp in layer.components:
                    # Check our alternate layer set-up agrees with theirs
                    components_bracket_layers = {
                        tuple(layer._bracket_axis_rules())
                        for layer in alternate_layers[master][comp.name]
                    }
                    if my_bracket_layers != components_bracket_layers:
                        # Find what we need to add
                        needed = components_bracket_layers - my_bracket_layers
                        if needed:
                            problematic_glyphs[(glyph_name, master)] |= needed

        if not problematic_glyphs:
            break

        # And now, fix the problem.
        for (glyph_name, master), needed_brackets in problematic_glyphs.items():
            my_bracket_layers = [
                tuple(layer._bracket_axis_rules())
                for layer in alternate_layers[master][glyph_name]
            ]
            if my_bracket_layers:
                # We have some bracket layers, but they're not the ones we
                # expect. Do the wrong thing, because doing the right thing
                # requires major investment.
                master_name = font.masters[master].name
                logger.warning(
                    f"Glyph {glyph_name} in master {master_name} has different "
                    "alternate layers to components that it uses. We don't "
                    "currently support this case, so some alternate layers will "
                    "not be applied. Consider fixing the source instead."
                )
            # Just copy the master layer for each thing we need.
            # Insert the new layers in a predictable, deterministic order based on
            # the bracket layers' axis values.
            for axis_rules in sorted(
                needed_brackets,
                key=lambda rules: tuple(
                    (
                        float("-inf") if r[0] is None else r[0],
                        float("inf") if r[1] is None else r[1],
                    )
                    for r in rules
                ),
            ):
                new_layer = synthesize_bracket_layer(
                    master_layers[master][glyph_name], axis_rules
                )
                font.glyphs[glyph_name].layers.append(new_layer)
                alternate_layers[master][glyph_name].append(new_layer)


def synthesize_bracket_layer(old_layer, axis_rules):
    new_layer = copy.copy(old_layer)  # We don't need a deep copy of everything
    new_layer.layerId = ""
    new_layer.associatedMasterId = old_layer.layerId

    if new_layer.parent.parent.format_version == 2:
        bottom, top = next(iter(axis_rules))
        if bottom is None:
            new_layer.name = old_layer.name + f" ]{top}]"
        elif top is None:
            new_layer.name = old_layer.name + f" [{bottom}]"
        else:
            raise AssertionError(f"either {bottom} or {top} must be None")
    else:
        new_layer.attributes = dict(
            new_layer.attributes
        )  # But we do need our own version of this
        new_layer.attributes["axisRules"] = [
            {"min": bottom, "max": top} for (bottom, top) in axis_rules
        ]

    assert tuple(new_layer._bracket_axis_rules()) == axis_rules
    return new_layer



import copy
from dataclasses import dataclass
from enum import IntEnum
import logging
import math

from fontTools.misc.bezierTools import (
    _alignment_transformation,
    calcCubicParameters,
    solveCubic,
    cubicPointAtT,
    linePointAtT,
    segmentPointAtT,
    splitCubic,
)
from fontTools.pens.reverseContourPen import ReverseContourPen
from fontTools.misc.roundTools import otRound
from fontTools.misc.transform import Transform
from ufo2ft.filters import BaseFilter
from ufoLib2.objects import Glyph

from glyphsLib.builder.constants import HINTS_LIB_KEY, SHAPE_ORDER_LIB_KEY


try:
    from math import dist
except ImportError:

    def dist(p1, p2):
        return math.sqrt((p1[0] - p2[0]) ** 2 + (p1[1] - p2[1]) ** 2)


logger = logging.getLogger(__name__)


class Alignment(IntEnum):
    OUTSTROKE = 0  # Glyphs calls this "left" alignment
    INSTROKE = 1  # Glyphs calls this "right" alignment
    MIDDLE = 2
    UNUSED = 3
    UNALIGNED = 4


# Lots of boring curve math stuff...


def otRoundNode(node):
    node.x, node.y = otRound(node.x), otRound(node.y)
    return node


# We often have Points (of some unknown UFO class), but fontTools
# math stuff needs tuples.
def as_tuples(pts):
    return [(pt.x, pt.y) for pt in pts]


def get_next_segment(path, index):
    seg = [path[index]]
    index = (index + 1) % len(path)
    seg.append(path[index])
    if not seg[-1].type:
        index = (index + 1) % len(path)
        seg.append(path[index])
        index = (index + 1) % len(path)
        seg.append(path[index])
    return seg


def get_previous_segment(path, index):
    seg = [path[index]]
    index = (index - 1) % len(path)
    seg.append(path[index])
    if not seg[-1].type:
        index = (index - 1) % len(path)
        seg.append(path[index])
        index = (index - 1) % len(path)
        seg.append(path[index])
    return list(reversed(seg))


def closest_point_on_segment(seg, pt):
    # Everything here is a tuple
    if len(seg) == 2:
        return closest_point_on_line(seg, pt)
    return closest_point_on_cubic(seg, pt)


def closest_point_on_line(seg, pt):
    a, b = seg
    a_to_b = (b[0] - a[0], b[1] - a[1])
    a_to_pt = (pt[0] - a[0], pt[1] - a[1])
    mag = a_to_b[0] ** 2 + a_to_b[1] ** 2
    if mag == 0:
        return a
    atp_dot_atb = a_to_pt[0] * a_to_b[0] + a_to_pt[1] * a_to_b[1]
    t = atp_dot_atb / mag
    return (a[0] + a_to_b[0] * t, a[1] + a_to_b[1] * t)


def closest_point_on_cubic(bez, pt, start=0.0, end=1.0, iterations=5, slices=5):
    tick = (end - start) / slices
    best = 0
    best_dist = float("inf")
    t = start
    best_pt = pt
    while t < end:
        this_pt = cubicPointAtT(*bez, t)
        current_distance = dist(this_pt, pt)
        if current_distance <= best_dist:
            best_dist = current_distance
            best = t
            best_pt = this_pt
        t += tick
    if iterations < 1:
        return best_pt
    return closest_point_on_cubic(
        bez,
        pt,
        start=max(best - tick, 0),
        end=min(best + tick, 1),
        iterations=iterations - 1,
        slices=slices,
    )


def unbounded_seg_seg_intersection(seg1, seg2):
    if len(seg1) == 2 and len(seg2) == 2:
        aligned_seg1 = _alignment_transformation(seg1).transformPoints(seg2)
        if not math.isclose(aligned_seg1[0][1], aligned_seg1[1][1]):
            t = aligned_seg1[0][1] / (aligned_seg1[0][1] - aligned_seg1[1][1])
            return linePointAtT(*seg2, t)
        elif not math.isclose(aligned_seg1[0][0], aligned_seg1[1][0]):
            t = aligned_seg1[0][0] / (aligned_seg1[0][0] - aligned_seg1[1][0])
            return linePointAtT(*seg2, t)
        else:
            return None
    if len(seg1) == 4 and len(seg2) == 2:
        curve, line = seg1, seg2
    elif len(seg1) == 2 and len(seg2) == 4:
        line, curve = seg1, seg2
    aligned_curve = _alignment_transformation(line).transformPoints(curve)
    a, b, c, d = calcCubicParameters(*aligned_curve)
    intersections = solveCubic(a[1], b[1], c[1], d[1])
    real_intersections = [t for t in intersections if t >= 0 and t <= 1]
    if real_intersections:
        return cubicPointAtT(*curve, real_intersections[0])
    return None  # Needs bending


def point_on_seg_at_distance(seg, distance):
    aligned_seg = _alignment_transformation(seg).transformPoints(seg)
    if len(aligned_seg) == 4:
        a, b, c, d = calcCubicParameters(*aligned_seg)
        solutions = solveCubic(a[0], b[0], c[0], d[0] - (aligned_seg[0][0] + distance))
        solutions = sorted(t for t in solutions if 0 <= t < 1)
        if not solutions:
            return None
        return solutions[0]
    else:
        start, end = aligned_seg
        if math.isclose(end[0], start[0]):
            if math.isclose(end[1], start[1]):
                return 0
            return distance / (end[1] - start[1])
        else:
            return distance / (end[0] - start[0])


def split_cubic_at_point(seg, point, inward=True):
    # There's a horrible edge case here where the curve wraps around and
    # the ray hits twice, but I'm not worrying about it.
    if inward:
        new_cubic_1 = splitCubic(*seg, point[0], False)[0]
        new_cubic_2 = splitCubic(*seg, point[1], True)[0]
    else:
        new_cubic_1 = splitCubic(*seg, point[0], False)[-1]
        new_cubic_2 = splitCubic(*seg, point[1], True)[-1]
    if dist(new_cubic_1[-1], point) < dist(new_cubic_2[-1], point):
        return new_cubic_1
    else:
        return new_cubic_2


# Using a class here is mild overkill but it allows us to store
# the information about the component in a slightly more readable
# manner.
@dataclass
class CornerComponentApplier:
    corner_name: str
    glyph_name: str
    alignment: Alignment
    glyph: object
    path_index: int
    corner_path: object
    other_paths: list
    target_node: object
    target_node_ix: int = None
    origin: (int, int) = (0, 0)
    effective_start: (int, int) = None
    effective_end: (int, int) = None
    scale: (int, int) = None
    left_x: int = 0
    right_x: int = 0
    outstroke_intersection_point: (int, int) = None

    def fail(self, msg, hard=True):
        full_msg = f"{msg} (corner {self.corner_name} in {self.glyph_name})"
        if hard:
            raise ValueError(full_msg)
        else:
            logger.error(full_msg)

    @property
    def instroke(self):
        return get_previous_segment(self.path, self.target_node_ix)

    @property
    def outstroke(self):
        return get_next_segment(self.path, self.target_node_ix)

    @property
    def first_seg(self):
        return get_next_segment(self.corner_path, 0)

    @property
    def last_seg(self):
        return get_previous_segment(self.corner_path, len(self.corner_path) - 1)

    def apply(self):
        self.path = self.glyph[self.path_index]
        # Find where the target node lines in this path. This may have
        # changed, if we have applied a corner component in this path
        # already.
        for ix, node in enumerate(self.path):
            if node == self.target_node:
                self.target_node_ix = ix
        if self.target_node_ix is None:
            self.fail("Lost track of where the corner should be applied")

        if self.corner_path[0].x != self.origin[0]:
            self.fail(
                "Can't deal with offset instrokes yet; start corner components on axis",
                hard=False,
            )

        # This is for handling the left and right anchors and doesn't
        # quite work yet
        self.determine_start_and_end_vectors()

        # Align all paths to the "origin" anchor.
        for path in [self.corner_path] + self.other_paths:
            for pt in path:
                pt.x, pt.y = pt.x - self.origin[0], pt.y - self.origin[1]

        # Apply scaling. We are considered "flipped" if one or other
        # of the scale factors is negative, but not both. Being flipped
        # means that the corner path gets applied backwards.
        self.flipped = False
        if self.scale is not None:
            self.flipped = (self.scale[0] * self.scale[1]) < 0
            self.scale_paths()

        # Align and rotate the corner paths so that they fit onto
        # the host path
        (
            instroke_intersection_point,
            outstroke_intersection_point,
            correction,
        ) = self.align_my_path_to_main_path()

        # Keep hold of the original outstroke segment. Fitting the
        # instroke to the corner component will change the position
        # of the target node (since it's at the end of that segment)
        # so we need to recover it later.
        original_outstroke = as_tuples(self.outstroke)

        # If we are not aligned to the instroke, we need to re-fit the
        # instroke based on where we put the corner component, and
        # potentially stretch the corner component so that it meets the
        # instroke.
        if self.alignment != Alignment.INSTROKE and correction:
            instroke_intersection_point = self.recompute_instroke_intersection_point()
            # The instroke of the corner path may need stretching to fit...
            if len(self.first_seg) == 4:
                self.stretch_first_seg_to_fit(instroke_intersection_point)
        self.split_instroke(instroke_intersection_point)

        # Now we insert the aligned and rotated corner path into the host
        self.path[self.target_node_ix + 1 : self.target_node_ix + 1] = [
            otRoundNode(node) for node in self.corner_path[1:]
        ]

        # And fix up the outstroke
        outstroke_intersection_point = self.recompute_outstroke_intersection_point(
            original_outstroke
        )
        self.fixup_outstroke(original_outstroke, outstroke_intersection_point)

        # Last of all, if there are other paths in the corner component,
        # they just get copied into the glyph.
        self.insert_other_paths()

    def determine_start_and_end_vectors(self):
        # Left and right anchors provide an additional offset, depending
        # on their relationship with the start/end of the first/last points
        # of the corner seg
        if not self.effective_start:
            self.effective_start = (0, 0)
        else:
            self.effective_start = (
                self.corner_path[0].x - self.effective_start[0],
                self.corner_path[0].y - self.effective_start[1],
            )
            self.fail(
                "left and right anchors to corner components are"
                " not currently supported",
                hard=False,
            )

        if not self.effective_end:
            self.effective_end = (0, 0)
        else:
            self.effective_end = (
                self.corner_path[-1].x - self.effective_end[0],
                self.corner_path[-1].y - self.effective_end[1],
            )
            self.fail(
                "left and right anchors to corner components are"
                " not currently supported",
                hard=False,
            )

    def scale_paths(self):
        scaling = Transform().scale(*self.scale)
        for path in [self.corner_path] + self.other_paths:
            for pt in path:
                pt.x, pt.y = scaling.transformPoint((pt.x, pt.y))

    def align_my_path_to_main_path(self):
        # Work out my rotation (1): Rotation occurring due to corner paths
        angle = math.atan2(-self.corner_path[-1].y, self.corner_path[-1].x)

        # Work out my rotation (2): Rotation occurring due to host paths
        if self.flipped:
            angle += math.radians(90)

            self.reverse_corner_path()

        # To align along the outstroke, work out how much the end of the
        # corner pokes out, then find a point on the curve that distance
        # away. Use that as the vector
        distance = self.last_seg[-1].y if self.flipped else self.last_seg[-1].x
        t = point_on_seg_at_distance(as_tuples(self.outstroke), distance)
        outstroke_intersection_point = segmentPointAtT(as_tuples(self.outstroke), t)
        outstroke_angle = math.atan2(
            outstroke_intersection_point[1] - self.target_node.y,
            outstroke_intersection_point[0] - self.target_node.x,
        )

        # And the same for the instroke, determined by the Y value of
        # the first point on the corner component
        distance = -self.first_seg[0].x if self.flipped else self.first_seg[0].y
        t2 = point_on_seg_at_distance(as_tuples(self.instroke), distance)
        instroke_intersection_point = segmentPointAtT(
            as_tuples(reversed(self.instroke)), t2
        )
        correction = not (math.isclose(t2, 0.0) or math.isclose(t2, 1.0))
        instroke_angle = math.atan2(
            self.target_node.y - instroke_intersection_point[1],
            self.target_node.x - instroke_intersection_point[0],
        )
        instroke_angle += math.radians(90)

        if self.alignment == Alignment.OUTSTROKE:
            angle += outstroke_angle
        elif self.alignment == Alignment.INSTROKE:
            angle += instroke_angle
        elif self.alignment == Alignment.MIDDLE:
            angle += (instroke_angle + outstroke_angle) / 2
        else:  # Unaligned, do nothing
            pass

        # Rotate the paths around the origin and then align them
        # so that the origin of the corner starts on the target node
        rot = Transform().rotate(angle)
        translation = Transform().translate(
            self.target_node.x + self.effective_start[0], self.target_node.y
        )
        for path in [self.corner_path] + self.other_paths:
            for pt in path:
                pt.x, pt.y = translation.transform(rot).transformPoint((pt.x, pt.y))

        return instroke_intersection_point, outstroke_intersection_point, correction

    def recompute_instroke_intersection_point(self):
        return unbounded_seg_seg_intersection(
            as_tuples(self.first_seg[0:2]), as_tuples(self.instroke)
        )

    def recompute_outstroke_intersection_point(self, original_outstroke):
        if self.flipped:
            # Project it
            return unbounded_seg_seg_intersection(
                as_tuples(self.last_seg), original_outstroke
            )

        # Bend it
        return closest_point_on_segment(
            original_outstroke,
            (self.corner_path[-1].x, self.corner_path[-1].y),
        )

    def split_instroke(self, intersection):
        if len(self.instroke) == 2:
            # Splitting a line is easy...
            (
                self.path[self.target_node_ix].x,
                self.path[self.target_node_ix].y,
            ) = otRound(intersection[0]), otRound(intersection[1])
        else:
            new_cubic = split_cubic_at_point(
                as_tuples(self.instroke), intersection, inward=True
            )
            for new_pt, old in zip(new_cubic, self.instroke):
                old.x, old.y = otRound(new_pt[0]), otRound(new_pt[1])

    def fixup_outstroke(self, original_outstroke, intersection):
        # Split the outstroke at the nearest point to the intersection.
        # The outstroke has moved now, since we have inserted the path
        outstroke = get_next_segment(
            self.path,
            (self.target_node_ix + len(self.corner_path) - 1) % len(self.path),
        )

        if not intersection:
            # Something's probably wrong...
            return

        if len(outstroke) == 2:
            (outstroke[0].x, outstroke[0].y) = otRound(intersection[0]), otRound(
                intersection[1]
            )
        else:
            new_cubic = split_cubic_at_point(
                original_outstroke, intersection, inward=False
            )
            for new_pt, old in zip(new_cubic, outstroke):
                old.x, old.y = otRound(new_pt[0]), otRound(new_pt[1])

    def stretch_first_seg_to_fit(self, intersection):
        delta = (
            intersection[0] - self.first_seg[0].x,
            intersection[1] - self.first_seg[0].y,
        )
        self.first_seg[1].x += delta[0]
        self.first_seg[1].y += delta[1]

    def reverse_corner_path(self):
        new_glyph = Glyph()
        self.corner_path.draw(ReverseContourPen(new_glyph.getPen()))
        self.corner_path[:] = new_glyph[0]

    def insert_other_paths(self):
        for path in self.other_paths:
            for node in path:
                otRoundNode(node)
            self.glyph.contours.append(path)


class CornerComponentsFilter(BaseFilter):
    def filter(self, glyph):
        if not len(glyph) or HINTS_LIB_KEY not in glyph.lib:
            return False

        corner_components = [
            hint
            for hint in glyph.lib[HINTS_LIB_KEY]
            if hint.get("type").upper() == "CORNER"
        ]

        if not corner_components:
            return False

        todo_list = []

        for glyphs_cc in corner_components:
            shape_index, node_idx = glyphs_cc["origin"]
            path_indices = {}
            if SHAPE_ORDER_LIB_KEY in glyph.lib:
                # Map between shape index and path index
                for ix, sign in enumerate(glyph.lib[SHAPE_ORDER_LIB_KEY]):
                    if sign == "P":
                        path_indices[ix] = len(path_indices.keys())
                if shape_index not in path_indices:
                    raise ValueError(
                        f"Could not find shape number {shape_index} in {glyph.name}"
                    )
            path_idx = path_indices.get(shape_index, shape_index)

            # We use font, not .glyphSet here because corner components
            # aren't normally exported
            if glyphs_cc["name"] not in self.context.font:
                logger.warn(
                    "Corner component %s in %s not found",
                    glyphs_cc["name"],
                    glyph.name,
                )
                continue
            layer = self.context.font[glyphs_cc["name"]]
            cc_anchor_dict = {anchor.name: anchor for anchor in layer.anchors}
            if "origin" in cc_anchor_dict:
                cc_origin = cc_anchor_dict["origin"].x, cc_anchor_dict["origin"].y
            else:
                cc_origin = (0, 0)

            if "left" in cc_anchor_dict:
                cc_left = cc_anchor_dict["left"].x, cc_anchor_dict["left"].y
            else:
                cc_left = None

            if "right" in cc_anchor_dict:
                cc_right = cc_anchor_dict["right"].x, cc_anchor_dict["right"].y
            else:
                cc_right = None

            corner_path = copy.deepcopy(layer[0])
            other_paths = [copy.deepcopy(path) for path in layer[1:]]

            cc = CornerComponentApplier(
                glyph_name=glyph.name,
                corner_name=glyphs_cc["name"],
                alignment=Alignment(glyphs_cc.get("options", 0)),
                scale=glyphs_cc.get("scale"),
                corner_path=corner_path,
                other_paths=other_paths,
                path_index=path_idx,
                glyph=glyph,
                origin=cc_origin,
                effective_start=cc_left,
                effective_end=cc_right,
                # We pass in the current starting node, because its
                # position may change if we apply more than one corner.
                target_node=glyph[path_idx][(node_idx + 1) % len(glyph[path_idx])],
            )
            todo_list.append(cc)

        for cc in todo_list:
            cc.apply()

        return True
